# Copyright 2021 Adobe. All rights reserved.

"""
Internal representation of a T2 CharString glyph with hints
"""

import threading
import operator
from copy import deepcopy
from math import sqrt
from collections import defaultdict
from builtins import tuple as _tuple
from typing import Iterator, Any, Callable
import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from fontTools.misc.bezierTools import (
    solveQuadratic,
    solveCubic,
    calcCubicParameters,
    splitCubicAtT,
    segmentPointAtT,
    approximateCubicArcLength,
)
from fontTools.pens.basePen import BasePen

import logging

from . import Number

log = logging.getLogger(__name__)


def norm_float(value: int | float) -> int | float:
    """Converts a float (whose decimal part is zero) to integer"""
    if isinstance(value, float):
        value = round(value, 4)
        if value.is_integer():
            return int(value)
        return value
    return value


def feq(a: int | float, b: int | float, factor: float = 1.52e-5) -> bool:
    """Returns True if a and b are close enough to be considered equal"""
    return abs(a - b) < factor


def fne(a: int | float, b: int | float, factor: float = 1.52e-5) -> bool:
    """Returns True if a and b are not close enough to be considered equal"""
    return abs(a - b) >= factor


class pt(tuple):
    """A 2-tuple representing a point in 2D space"""
    __slots__ = ()
    tl = threading.local()
    tl.align = None

    @classmethod
    def setAlign(cls, vertical: bool = False) -> None:
        """
        Class-level method to control the value of properties a and o.
        "a" is meant so suggest "aligned" and "o" is meant to suggest
        "opposite".

        When called with vertical==False (the default):
            a will be equivalent to x
            o will be equivalent to y
        When called with vertical==True
            a will be equivalent to y
            o will be equivalent to x

        Note that the internal align variable is thread-specific
        """
        if vertical:
            cls.tl.align = 2
        else:
            cls.tl.align = 1

    @classmethod
    def clearAlign(cls) -> None:
        """
        Class-level method to unset the internal align variable
        so that accessing properties a or o will result in an error
        """
        cls.tl.align = None

    def __new__(
        cls,
        x: int | float | tuple[int, int] = 0,
        y: int | float = 0,
        roundCoords: bool = False
    ):
        """
        Creates a new pt object initialied with x and y.

        If roundCoords is True the values are rounded before storing
        """
        if isinstance(x, tuple):
            y = x[1]
            x = x[0]
        if roundCoords:
            x = round(x)
            y = round(y)
        return _tuple.__new__(cls, (x, y))

    @property
    def x(self) -> Number:
        return self[0]

    @property
    def y(self) -> Number:
        return self[1]

    @property
    def a(self) -> int:
        """See note in setAlign"""
        if self.tl.align == 1:
            return self[0]
        elif self.tl.align == 2:
            return self[1]
        else:
            raise RuntimeError("glyphData.pt property a used without " +
                               "setting align")

    @property
    def o(self) -> int:
        """See note in setAlign"""
        if self.tl.align == 1:
            return self[1]
        elif self.tl.align == 2:
            return self[0]
        else:
            raise RuntimeError("glyphData.pt property o used without " +
                               "setting align")

    def ao(self):
        """See note in setAlign"""
        if self.tl.align == 1:
            return (self.x, self.y)
        elif self.tl.align == 2:
            return (self.y, self.x)
        else:
            raise RuntimeError("glyphData.pt method ao used without " +
                               "setting align")

    def norm_float(self):
        return pt(norm_float(self[0]), norm_float(self[1]))

    # for two pts
    def __add__(self, other):
        """
        Returns a new pt object representing the sum of respective
        values of the two arguments
        """
        if not isinstance(other, pt):
            raise TypeError('Both arguments to pt.__add__ must be pts')
        return pt(self[0] + other[0], self[1] + other[1])

    def __sub__(self, other: "pt") -> "pt":
        """
        Returns a new pt object representing the difference of respective
        values of the two arguments
        """
        if not isinstance(other, pt):
            raise TypeError('Both arguments to pt.__sub__ must be pts')
        return pt(self[0] - other[0], self[1] - other[1])

    def avg(self, other):
        """
        Returns a new pt object representing the average of this pt object
        with the argument
        """
        if not isinstance(other, pt):
            raise TypeError('Both arguments to pt.avg must be pts')
        return pt((self[0] + other[0]) * 0.5, (self[1] + other[1]) * 0.5)

    def dot(self, other: "pt") -> int:
        """
        Returns a numeric value representing the dot product of this
        pt object with the argument
        """
        if not isinstance(other, pt):
            raise TypeError('Both arguments to pt.dot must be pts')
        return self[0] * other[0] + self[1] * other[1]

    def scross(self, other):
        """
        Returns a numeric value representing the cross product of this
        pt object with the argument
        """
        if not isinstance(other, pt):
            raise TypeError('Both arguments to pt.dot must be pts')
        return other[1] * self[0] - other[0] * self[1]

    def distsq(self, other):
        """
        Returns a numerical value representing the squared distance
        between this pt object and the argument
        """
        if not isinstance(other, pt):
            raise TypeError('Both arguments to pt.distsq must be pts')
        dx = self[0] - other[0]
        dy = self[1] - other[1]
        return dx * dx + dy * dy

    def a_dist(self, other: "pt") -> int:
        """
        Returns a numerical value representing the distance between this
        pt object and the argument in the "a" dimension
        """
        return abs((self - other).a)

    def o_dist(self, other: "pt") -> int:
        """
        Returns a numerical value representing the distance between this
        pt object and the argument in the "o" dimension
        """
        return abs((self - other).o)

    def normsq(self) -> int:
        """Returns the squared magnitude of the pt (treated as a vector)"""
        return self[0] * self[0] + self[1] * self[1]

    def abs(self) -> "pt":
        """
        Returns a new pt object with the absolute values of the coordinates
        """
        return pt(abs(self[0]), abs(self[1]))

    def round(self, dec: int = 0):
        """Returns a new pt object with rounded coordinate values"""
        return pt(round(self[0], dec), round(self[1], dec))

    # for pt and number
    def __mul__(self, other):
        """
        Returns a new pt object with this object's coordinates multiplied by
        a scalar value
        """
        if not isinstance(other, (int, float)):
            raise TypeError('One argument to pt.__mul__ must be a scalar ' +
                            'number')
        return pt(self[0] * other, self[1] * other)

    def __rmul__(self, other):
        """Same as __mul__ for right-multiplication"""
        if not isinstance(other, (int, float)):
            raise TypeError('One argument to pt.__rmul__ must be a scalar ' +
                            'number')
        return pt(self[0] * other, self[1] * other)

    def __eq__(
        self, other: object, factor: float = 1.52e-5
    ) -> bool:  # type: ignore[override]
        """Returns True if each coordinate is feq to that of the argument"""
        if not isinstance(other, pt):
            return NotImplemented
        return (feq(self[0], other[0], factor) and
                feq(self[1], other[1], factor))

    def eq_exact(self, other: "pt") -> bool:
        """Returns True if each coordinate is equal to that of the argument"""
        return self[0] == other[0] and self[1] == other[1]


class stem(tuple):
    """
    A 2-tuple representing a stem hint.

    self[0] is the bottom/left coordinate
    self[1] is the top/right coordinate

    If self[1] is less than self[0] the stem represents a ghost hint
    and the difference should be 20 or 21 points
    """
    BandMargin = 30
    __slots__ = ()

    def __new__(cls, lb: Number = 0, rt: Number = 0):
        if isinstance(lb, tuple):
            return _tuple.__new__(cls, lb)
        return _tuple.__new__(cls, (lb, rt))

    @property
    def lb(self) -> int | float:
        """The left or bottom value, depending on stem alignment"""
        return self[0]

    @property
    def rt(self) -> int | float:
        """The right or top value, depending on stem alignment"""
        return self[1]

    def __str__(self) -> str:
        """Returns a string representation of the stem"""
        isgh = self.isGhost()
        isbad = self.isBad()
        if not isgh and not isbad:
            return "({0} -> {1})".format(self.lb, self.rt)
        elif isgh == "high":
            return "({0} (high ghost))".format(self.rt)
        elif isgh == "low":
            return "({0} (low ghost))".format(self.rt)
        else:
            return "(bad stem values {0}, {1})".format(self.lb, self.rt)

    def isGhost(self, doBool: bool = False) -> str | bool:
        """Returns True if the stem is a ghost hint"""
        width = self.rt - self.lb
        if width == -20:   # from topGhst in ac.h
            return "high" if not doBool else True
        if width == -21:   # from botGhst in ac.h
            return "low" if not doBool else True
        return False

    def ghostCompat(self, other: "stem") -> bool:
        if self.isGhost(True) == other.isGhost(True):
            return False
        if self.isGhost() == 'high' or other.isGhost() == 'high':
            return self.lb == other.rt
        else:
            return self.lb == other.lb

    def isBad(self) -> bool:
        """Returns True if the stem is malformed"""
        return not self.isGhost() and (self.rt - self.lb) < 0

    def relVals(self, last: "stem | None" = None) -> tuple[Number, Number]:
        """
        Returns a tuple of "relative" stem values (start relative to
        the passed last stem, then width) appropriate for
        vstem/hstem/vstemhm/hstemhm output
        """
        if last:
            l = last.rt
        else:
            l = 0
        return (norm_float(self.lb - l), norm_float(self.rt - self.lb))

    def UFOVals(self) -> tuple[Number, Number]:
        """Returns a tuple of stem values appropriate for UFO output"""
        return (self.lb, self.rt - self.lb)

    def overlaps(self, other: "stem") -> bool:
        """
        Returns True if this stem is within BandMargin of overlapping the
        passed stem
        """
        lloc, uloc = self.lb, self.rt
        sig = self.isGhost()
        # c.f. page 22 of Adobe TN #5177 "The Type 2 Charstring Format"
        if sig == 'high':
            uloc = lloc
        elif sig == 'low':
            lloc = uloc
        olloc, ouloc = other.lb, other.rt
        oig = other.isGhost()
        if oig == 'high':
            ouloc = olloc
        elif oig == 'low':
            olloc = ouloc
        uloc += stem.BandMargin
        lloc -= stem.BandMargin
        return olloc <= uloc and ouloc >= lloc

    def distance(self, loc: int) -> int | float:
        """
        Returns the distance between this stem and the passed location,
        which is zero if the location falls within the stem
        """
        lloc, uloc = self.lb, self.rt
        sig = self.isGhost()
        # c.f. page 22 of Adobe TN #5177 "The Type 2 Charstring Format"
        if sig == 'high':
            uloc = lloc
        elif sig == 'low':
            lloc = uloc
        if loc >= lloc and loc <= uloc:
            return 0
        return min(abs(lloc - loc), abs(uloc - loc))


class boundsState:
    """
    Calculates and stores the bounds of a pathElement (spline) and the point
    locations that define the boundaries.
    """
    def __init__(self, c) -> None:
        """
        Initialize the object with the passed pathElement and calculate the
        bounds
        """
        # Start with the bounds of the line betwen the endpoints
        self.lb = self.linearBounds(c)
        if c.is_line:
            self.bounds = self.lb
        else:
            # For curves, add the control points into the calcuation.
            # If they exceed the linear bounds calculate the proper
            # curve bounds
            cpb = deepcopy(self.lb)
            self.mergePt(cpb, c.cs, .25, doExt=False)  # t value isn't used
            self.mergePt(cpb, c.ce, .75, doExt=False)  # t value isn't used
            if self.lb != cpb:
                self.calcCurveBounds(c)
            else:
                self.bounds = self.lb

    def mergePt(self, b, p, t, doExt: bool = True) -> None:
        """
        Add the passed point into the bounds as a potential extreme.

        If it is an extreme
            store the point at the appropriate extpts subscripts
            store the t value at the same tmap subscripts
        """
        for i, cmp_o in enumerate([operator.lt, operator.gt]):
            for j in range(2):
                if cmp_o(p[j], b[i][j]):
                    b[i][j] = p[j]
                    if doExt:
                        self.extpts[i][j] = p
                        self.tmap[i][j] = t

    def linearBounds(self, c):
        """
        Calculate the bounds of the line betwen the start and end points of
        the passed pathElement.
        """
        self.tmap = [[0, 0], [0, 0]]
        self.extpts = [[c.s, c.s], [c.s, c.s]]
        lb = [[*c.s], [*c.s]]
        self.mergePt(lb, c.e, 1)
        return lb

    def calcCurveBounds(self, pe) -> None:
        """
        Calculate the bounds of the passed path element relative to the
        already-calculated linear bounds
        """
        curveb = deepcopy(self.lb)
        a, b, c, d = pe.cubicParameters()
        roots = []
        for dim in range(2):
            roots += [t for t in solveQuadratic(a[dim] * 3, b[dim] * 2, c[dim])
                      if 0 <= t < 1]

        for t in roots:
            p = pt(a[0] * t * t * t + b[0] * t * t + c[0] * t + d[0],
                   a[1] * t * t * t + b[1] * t * t + c[1] * t + d[1])
            self.mergePt(curveb, p, t)

        self.bounds = curveb

    def farthestExtreme(self, doY: bool = False):
        """
        Returns the location, defining point, and t value for the
        bound farthest from the linear bounds in the dimension selected
        with doY. If the linear bounds are the curve bounds returns
        0, None, None

        The fourth return value is False if the defining point's location
        is less than the linear bound and True if it is greater
        """
        idx = 1 if doY else 0
        d = [abs(self.lb[i][idx] - self.bounds[i][idx]) for i in range(2)]
        if d[0] > d[1]:
            return d[0], self.extpts[0][idx], self.tmap[0][idx], False
        elif fne(d[1], 0):
            return d[1], self.extpts[1][idx], self.tmap[1][idx], True
        else:
            return 0, None, None, None

    def intersects(self, other, margin: int = 0):
        """
        Returns True if the bounds of this object are within those of
        the argument
        """
        return (self.bounds[0][0] <= other.bounds[1][0] + margin and
                self.bounds[1][0] + margin >= other.bounds[0][0] and
                self.bounds[0][1] <= other.bounds[1][1] + margin and
                self.bounds[1][1] + margin >= other.bounds[0][1])


class pathBoundsState:
    """
    Calculates and stores the bounds of a glyphData object (path) and
    the pathElements (splines) that define the boundaries.
    """
    def __init__(self, pe) -> None:
        """Initialize the bounds with those of a single pathElement"""
        self.bounds = pe.getBounds().bounds
        self.extpes = [[pe, pe], [pe, pe]]

    def merge(self, other) -> None:
        """Merge this pathBoundsState object with the bounds of another"""
        for i, cmp_o in enumerate([operator.lt, operator.gt]):
            for j in range(2):
                if cmp_o(other.bounds[i][j], self.bounds[i][j]):
                    self.bounds[i][j] = other.bounds[i][j]
                    self.extpes[i][j] = other.extpes[i][j]

    def within(self, other):
        """
        Returns True if the bounds of this object are within those of
        the argument
        """
        return (self.bounds[0][0] >= other.bounds[0][0] and
                self.bounds[0][1] >= other.bounds[0][1] and
                self.bounds[1][0] <= other.bounds[1][0] and
                self.bounds[1][1] <= other.bounds[1][1])


class pathElement:
    """
    Stores the coordinates of a spline (line or curve) and
        hintmask values to add directly before the spline
        Whether the spline is the first or second part of a flex hint
        a boundsState object for the spline
        The position (subpath, offset) of the spline in the glyphData path

        self.s is the first point, self.e is the last.
        If the spline is a curve self.cs is the first control point and
        self.ce is the second.

        When segment_sub is not None it must either be a list or an
        integer. If it is a list then that list must contain pathElements,
        representing the same path (in order). These will be used in place
        of the pathElement during segment generation. Each of these
        substitution pathElements must have segment_sub equal to the
        offset in the parent's segment_sub list.
    """
    assocMatchFactor = 93
    tSlop = .005
    middleMult = 2

    s: pt
    e: pt
    cs: pt
    ce: pt

    def __init__(
        self,
        *args,
        is_close: bool = False,
        masks=None,
        flex: bool = False,
        position=None
    ) -> None:
        self.is_line = False
        self.is_close = is_close
        for p in args:
            if not isinstance(p, pt):
                raise TypeError('Positional arguments to ' +
                                'pathElement.__init__ must be pt objects')
        if len(args) == 2:
            self.is_line = True
            self.s = args[0]
            self.e = args[1]
        elif len(args) == 4:
            if is_close:
                raise TypeError('is_close must be False for path curves')
            self.s = args[0]
            self.cs = args[1]
            self.ce = args[2]
            self.e = args[3]
        else:
            raise TypeError('Wrong number of positional arguments to ' +
                            'pathElement.__init__')
        self.masks = masks
        self.flex = flex
        self.bounds: boundsState | None = None
        self.position: tuple[int, int] = position or (-1, -1)
        self.segment_sub: list["pathElement"] | int | None = None

    def getBounds(self) -> boundsState:
        """Returns the bounds object for the object, generating it if needed"""
        if self.bounds is not None:
            return self.bounds
        self.bounds = boundsState(self)
        return self.bounds

    def clearTempState(self) -> None:
        self.bounds = None
        self.segment_sub = None

    def isLine(self) -> bool:
        """Returns True if the spline is a line"""
        return self.is_line

    def isClose(self) -> bool:
        """Returns True if this pathElement implicitly closes a subpath"""
        return self.is_close

    def isStart(self) -> bool:
        """Returns True if this pathElement starts a subpath"""
        assert self.position is not None
        return self.position[1] == 0

    def isTiny(self) -> bool:
        """
        Returns True if the start and end points of the spline are within
        two em-units in both dimensions
        """
        d = (self.e - self.s).abs()
        return d.x < 2 and d.y < 2

    def isShort(self) -> bool:
        """
        Returns True if the start and end points of the spline are within
        about six em-units
        """
        d = (self.e - self.s).abs()
        mx, mn = sorted(tuple(d))
        return mx + mn * .336 < 6  # head.c IsShort

    def convertToLine(self) -> None:
        """
        If the pathElement is not already a line, make it one with the same
        start and end points
        """
        if self.is_line:
            return
        if self.flex is not None:
            log.error("Cannot convert flex-hinted curve to line: skipping.")
            return
        self.is_line = True
        del self.cs
        del self.ce
        self.bounds = None

    def convertToCurve(
        self,
        sRatio: float = .333333,
        eRatio=None,
        roundCoords: bool = False
    ) -> None:
        """
        If the pathElement is not already a curve, make it one. The control
        points are made colinear to preseve the shape. self.cs will be
        positioned at ratio sRatio from self.s and self.ce will be positioned
        at eRatio away from self.e
        """
        if not self.is_line:
            return
        if eRatio is None:
            eRatio = sRatio
        self.is_line = False
        # XXX deal with roundCoords properly
        self.cs = self.s * (1 - sRatio) + self.e * sRatio
        self.ce = self.s * eRatio + self.e * (1 - eRatio)

    def clearHints(self, doVert: bool = False) -> None:
        """Clear the vertical or horizontal masks, if any"""
        if doVert and self.masks is not None:
            self.masks = [self.masks[0], None] if self.masks[0] else None
        elif not doVert and self.masks is not None:
            self.masks = [None, self.masks[1]] if self.masks[1] else None

    def cubicParameters(self) -> tuple[
        tuple[float, float],
        tuple[float, float],
        tuple[float, float],
        tuple[float, float],
    ]:
        """Returns the fontTools cubic parameters for this pathElement"""
        return calcCubicParameters(self.s, self.cs, self.ce, self.e)

    def getAssocFactor(self, loose: bool = False):
        if self.is_line:
            l = sqrt(self.s.distsq(self.e))
        else:
            l = approximateCubicArcLength(self.s, self.cs, self.ce, self.e)
        return l / (self.assocMatchFactor / 2
                    if loose else self.assocMatchFactor)

    def containsPoint(self, p, factor, returnT: bool = False):
        if self.is_line:
            # We sometimes want t anyway, so why not?
            ds = sqrt(self.s.distsq(self.e))
            d1 = sqrt(self.s.distsq(p))
            d2 = sqrt(self.e.distsq(p))
            iseq = feq(ds, d1 + d2, factor)
            if not returnT:
                return iseq
            else:
                return iseq, d1 / ds
        a, b, c, d = self.cubicParameters()
        if abs(self.s.x - self.e.x) > abs(self.s.y - self.e.y):
            i, j = 0, 1
        else:
            i, j = 1, 0
        tl = solveCubic(a[i], b[i], c[i], d[i] - p[i])
        sols = [(a[j] * t * t * t + b[j] * t * t + c[j] * t + d[j], t)
                for t in tl if -.1 <= t <= 1.1]
        tsols = [t for v, t in sols if feq(v, p[j], factor)]
        if not returnT:
            return len(tsols) > 0
        if len(tsols) > 0:
            return True, tsols[0]
        else:
            return False, 0

    def slopePoint(self, t):
        """
        Returns the point definiing the slope of the pathElement
        (relative to the on-curve point) at t==0 or t==1
        """
        if t == 1:
            if self.is_line:
                return self.s
            if self.e == self.ce:
                return self.cs
            return self.ce
        else:
            if self.is_line:
                return self.e
            if self.s == self.cs:
                return self.ce
            return self.cs

    def __deepcopy__(self, memo):
        """Don't deepcopy pathElement objects"""
        return self

    @staticmethod
    def stemBytes(masks):
        """Calculate bytes corresponding to a (boolean array) hintmask"""
        t = masks[0] + masks[1]
        lenb = len(t)
        lenB = (lenb + 7) // 8
        t += [False for i in range(lenB * 8 - lenb)]
        return int(''.join(('1' if i else '0' for i in t)),
                   2).to_bytes(lenB, byteorder='big')

    def relVals(self):
        """
        Return relative coordinates appropriate for an rLineTo or
        rCurveTo T2 operator
        """
        if self.is_line:
            r = self.e - self.s
            return [*r]
        else:
            r1 = self.cs - self.s
            r2 = self.ce - self.cs
            r3 = self.e - self.ce
            return [*r1, *r2, *r3]

    def T2(self, is_start: tuple[pt, Self] | None = None) -> tuple[list, pt]:
        """Returns an array of T2 operators corresponding to the pathElement"""
        prog = []

        if is_start:
            if is_start[1].masks:
                prog.extend(['hintmask',
                             pathElement.stemBytes(is_start[1].masks)])
            rmt = self.s - is_start[0]
            prog.extend([*rmt, 'rmoveto'])

        if self.masks and self.flex == 2:
            log.warning("Hintmask added in middle of flex hint: ignoring")
        elif self.masks and not self.is_close:
            prog.extend(['hintmask', pathElement.stemBytes(self.masks)])

        rv = self.relVals()

        after = self.e

        if self.is_close:
            after = self.s
        elif self.flex == 1:
            prog.extend(rv)
        elif self.flex == 2:
            prog.extend(rv)
            prog.extend([50, 'flex'])  # 50 from gDMin in C implementation
        elif not self.is_line:
            prog.extend(rv)
            prog.append('rrcurveto')
        else:
            prog.extend(rv)
            prog.append('rlineto')

        return prog, after

    def splitAtInflectionsForSegs(self) -> bool:
        a, b, c, d, = self.cubicParameters()
        t2c = 3 * (b[0] * a[1] - a[0] * b[1])
        t1c = 3 * (c[0] * a[1] - a[0] * c[1])
        t0c = (c[0] * b[1] - b[0] * c[1])
        sols = [t for t in solveQuadratic(t2c, t1c, t0c) if 0 < t < 1]
        if len(sols) > 0:
            self.segment_sub = []
            for s in splitCubicAtT(self.s, self.cs, self.ce, self.e, *sols):
                pts = [pt(tup) for tup in s]
                spe = pathElement(*pts)
                spe.position = self.position
                spe.segment_sub = len(self.segment_sub)
                self.segment_sub.append(spe)
            return True
        return False

    def splitAt(self, t: Number) -> Self:
        if self.is_line:
            pb = self.s + (self.e - self.s) * t
            ret = pathElement(pb, self.e, position=self.position,
                              is_close=self.is_close)
            self.e = pb
            self.is_close = False
            self.bounds = None
        else:
            s, n = splitCubicAtT(self.s, self.cs, self.ce, self.e, t)
            ptsn = [pt(p) for p in n]
            ret = pathElement(*ptsn, position=self.position,
                              is_close=self.is_close)
            assert s[0] == self.s and n[3] == ret.e
            self.cs = pt(s[1])
            self.ce = pt(s[2])
            self.e = pt(s[3])
            self.is_close = False
            self.bounds = None
        # The typing is correct but neither mypy nor pytype handle
        # self types well yet.
        return ret  # type: ignore

    def atT(self, t: Number) -> pt:
        return pt(segmentPointAtT(self.fonttoolsSegment(), t))

    def fonttoolsSegment(self) -> list[tuple[Number, Number]]:
        if self.is_line:
            return [(self.s[0], self.s[1]),
                    (self.e[0], self.e[1])]
        else:
            return [
                (self.s[0], self.s[1]),
                (self.cs[0], self.cs[1]),
                (self.ce[0], self.ce[1]),
                (self.e[0], self.e[1])]


class glyphData(BasePen):
    """Stores state corresponding to a T2 CharString"""

    def __init__(self, roundCoords: bool, name: str = '') -> None:
        super().__init__()
        self.roundCoords = roundCoords

        self.subpaths: list[Any] = []
        self.hstems: list[stem] = []
        self.vstems: list[stem] = []
        self.startmasks: list[Any] | None = None
        self.cntr: list[Any] = []
        self.name = name
        self.wdth: int | None = None

        self.is_hm: bool | None = None
        self.flex_count = 0
        self.lastcp = None

        self.nextmasks: list[Any] | None = None
        self.nextflex: int | None = None
        self.changed = False
        self.pathEdited = False
        self.boundsMap: dict[Any, Any] = {}

        self.hhs: object | None = None
        self.vhs: object | None = None

    # pen methods:

    def _moveTo(self, ptup: tuple[int, int]) -> None:
        """moveTo pen method"""
        self.lastcp = None
        self.subpaths.append([])

    def _lineTo(self, ptup: tuple[int, int]) -> None:
        """lineTo pen method"""
        self.lastcp = None
        curpt = pt(self._getCurrentPoint(), roundCoords=self.roundCoords)
        newpt = pt(ptup, roundCoords=self.roundCoords)
        self.subpaths[-1].append(pathElement(curpt, newpt,
                                             masks=self.getStemMasks(),
                                             flex=self.checkFlex(False),
                                             position=self.getPosition()))

    def _curveToOne(
        self,
        ptup1: tuple[int, int],
        ptup2: tuple[int, int],
        ptup3: tuple[int, int]
    ) -> None:
        """lineTo pen method"""
        self.lastcp = None
        curpt = pt(self._getCurrentPoint(), roundCoords=self.roundCoords)
        newpt1 = pt(ptup1, roundCoords=self.roundCoords)
        newpt2 = pt(ptup2, roundCoords=self.roundCoords)
        newpt3 = pt(ptup3, roundCoords=self.roundCoords)
        self.subpaths[-1].append(pathElement(curpt, newpt1, newpt2, newpt3,
                                             masks=self.getStemMasks(),
                                             flex=self.checkFlex(True),
                                             position=self.getPosition()))

    # closePath is a courtesy of the caller, not an instruction, so
    # we rely on its semantics here
    def _closePath(self) -> None:
        """closePath (courtesy) pen method"""
        curpt = pt(self._getCurrentPoint(), roundCoords=self.roundCoords)
        if len(self.subpaths[-1]) == 0:  # No content after moveTo
            self.subpaths.pop()
            return
        t = self.subpaths[-1][0].s
        if curpt.eq_exact(t):
            return
        self.subpaths[-1].append(pathElement(curpt, t,
                                             masks=self.getStemMasks(),
                                             flex=self.checkFlex(False),
                                             is_close=True,
                                             position=self.getPosition()))
        self.lastcp = self.subpaths[-1][-1]

    def getPosition(self) -> tuple[int, int]:
        """Returns position (subpath idx, offset) of next spline to be drawn"""
        return (len(self.subpaths) - 1, len(self.subpaths[-1]))

    # "hintpen" methods:
    def nextIsFlex(self) -> None:
        """quasi-pen method noting that next spline starts a flex hint"""
        self.flex_count += 1
        self.nextflex = 1

    def hStem(self, data, is_hm: bool) -> None:
        """
        quasi-pen method to pass horizontal stem data (in relative format)
        """
        if self.is_hm is not None and self.is_hm != is_hm:
            # XXX Should refactor so this prints as warning when
            # using existing hint data
            log.info("Horizontal stem hints mask setting does not match" +
                     " previous setting")
        self.is_hm = is_hm
        self.hstems.extend(self.toStems(data))

    def vStem(self, data, is_hm: bool | None) -> None:
        """quasi-pen method passing vertical stem data (in relative format)"""
        if is_hm is not None:
            if self.is_hm is not None and self.is_hm != is_hm:
                # XXX Should refactor so this prints as warning when
                # using existing hint data
                log.info("Vertical stem hints mask setting does not match" +
                         " previous setting")
            self.is_hm = is_hm
        self.vstems.extend(self.toStems(data))

    def hintmask(self, hhints, vhints) -> None:
        """quasi-pen method passing hintmask data"""
        if not self.is_hm:
            # XXX Should refactor so this prints as warning when
            # using existing hint data
            log.info("Hintmask found when stems weren't marked as having" +
                     " hintmasks")
            self.is_hm = True
        if len(self.subpaths) == 0:
            self.startmasks = [hhints, vhints]
        else:
            if not self.startmasks and not self.cntr:
                # XXX Should refactor so this prints as warning when
                # using existing hint data
                log.info("Initial hintmask missing in current glyph %s" %
                         self.name)
            # In the glyphdata format the end of a path is implicit in the
            # charstring but explicit in the subpath, while a moveto is
            # explicit in the charstring and implicit in the subpath. So
            # we record mid-stream hintmasks prior to movetos in the
            # prior "is_close" line. When writing out these will wind up
            # in the same place in the stream.
            if self.lastcp:
                self.lastcp.masks = [hhints, vhints]
            else:
                self.nextmasks = [hhints, vhints]

    def cntrmask(self, hhints, vhints) -> None:
        """quasi-pen method passing cntrmask data"""
        self.cntr.append([hhints, vhints])

    # width
    def setWidth(self, width: int | None) -> None:
        self.wdth = width

    def getWidth(self) -> int | None:
        return self.wdth

    # status
    def isEmpty(self) -> bool:
        """Returns true if there are no subpaths"""
        return not len(self.subpaths) > 0

    def hasFlex(self) -> bool:
        """Returns True if at least one curve pair is flex-hinted"""
        return self.flex_count > 0

    def hasHints(
        self,
        doVert: bool = False,
        both: bool = False,
        either: bool = False
    ) -> bool:
        """
        Returns True if there are hints of the parameter-specified type(s)
        """
        if both:
            return len(self.vstems) > 0 and len(self.hstems) > 0
        elif either:
            return len(self.vstems) > 0 or len(self.hstems) > 0
        elif doVert:
            return len(self.vstems) > 0
        else:
            return len(self.hstems) > 0

    def syncPositions(self) -> None:
        """
        Reset the pathElement.position tuples if the path has been edited
        """
        if self.pathEdited:
            for sp in range(len(self.subpaths)):
                for ofst in range(len(self.subpaths[sp])):
                    pe = self.subpaths[sp][ofst]
                    pe.position = (sp, ofst)
                    if isinstance(pe.segment_sub, list):
                        for i, spe in enumerate(pe.segment_sub):
                            spe.position = (sp, ofst)
                            spe.segment_sub = i

    def setPathEdited(self) -> None:
        self.pathEdited = True
        self.boundsMap = {}

    def getBounds(self, subpath=None):
        """
        Returns the bounds of the specified subpath, or of the whole
        path if subpath is None
        """
        b = self.boundsMap.get(subpath, None)
        if b is not None:
            return b
        if len(self.subpaths) == 0:
            return None
        if subpath is None:
            b = deepcopy(self.getBounds(0))
            for i in range(1, len(self.subpaths)):
                b.merge(self.getBounds(i))
            self.boundsMap[subpath] = b
            return b
        else:
            b = pathBoundsState(self.subpaths[subpath][0])
            for c in self.subpaths[subpath][1:]:
                b.merge(pathBoundsState(c))
            self.boundsMap[subpath] = b
            return b

    def T2(self, version: int = 1) -> list[int | str | float | bytes]:
        """Returns an array of T2 operators corresponding to the object"""
        prog = []

        if self.hstems:
            prog.extend(self.fromStems(self.hstems))
            prog.append('hstemhm' if self.is_hm else 'hstem')
        if self.vstems:
            prog.extend(self.fromStems(self.vstems))
            if not self.is_hm:
                prog.append('vstem')
            elif len(self.cntr) == 0 and self.startmasks is None:
                prog.append('vstemhm')

        for c in self.cntr:
            prog.extend(['cntrmask', pathElement.stemBytes(c)])
        if self.startmasks:
            prog.extend(['hintmask', pathElement.stemBytes(self.startmasks)])

        curpt = pt(0, 0)

        for c in self:
            is_start = None
            if c.isStart():
                is_start = (curpt, self.subpaths[c.position[0]][-1])
            c_t2, curpt = c.T2(is_start=is_start)
            prog.extend(c_t2)

        if version == 1:
            prog.append('endchar')
        return prog

    def drawPoints(self, pen: Any, ufoH: Callable | None = None) -> None:
        """
        Calls pointPen commands on pen to draw the glyph, optionally naming
        some points and building a library of hint annotations
        """
        doHints = ufoH is not None and (self.hasFlex() or
                                        self.hasHints(either=True))
        pln, pn = 0, None
        for s in self.subpaths:
            wrapi = len(s) - 1
            if wrapi < 0:
                continue
            w = s[wrapi]
            assert w.e == s[0].s
            if w.e == w.s:
                wrapi -= 1
                w = s[wrapi]
            pen.beginPath()
            wt = 'line' if w.isLine() else 'curve'
            if doHints:
                assert ufoH is not None
                pln, pn = ufoH(w, pln, True)
            pen.addPoint((w.e.x, w.e.y), segmentType=wt, name=pn)
            for i in range(0, wrapi):
                c = s[i]
                if doHints:
                    assert ufoH is not None
                    pln, pn = ufoH(c, pln)
                if c.isLine():
                    pen.addPoint((c.e.x, c.e.y), segmentType="line", name=pn)
                else:
                    pen.addPoint((c.cs.x, c.cs.y), name=pn)
                    pen.addPoint((c.ce.x, c.ce.y))
                    pen.addPoint((c.e.x, c.e.y), segmentType="curve")
            if not w.isLine():
                if doHints:
                    assert ufoH is not None
                    pln, pn = ufoH(w, pln)
                pen.addPoint((w.cs.x, w.cs.y), name=pn)
                pen.addPoint((w.ce.x, w.ce.y))
            pen.endPath()

    # XXX deal with or avoid reordering when preserving any hints
    def reorder(self, neworder) -> None:
        """Change the order of subpaths according to neworder"""
        log.debug("Reordering subpaths: %r" % neworder)
        spl = self.subpaths
        assert len(neworder) == len(spl)
        self.subpaths = [spl[i] for i in neworder]
        self.setPathEdited()

    def first(self):
        """Returns the first pathElement of the path"""
        if self.subpaths and self.subpaths[0]:
            return self.subpaths[0][0]
        return None

    def last(self):
        """Returns the last (implicit close) pathElement of the path"""
        if self.subpaths and self.subpaths[-1]:
            return self.subpaths[-1][-1]
        return None

    def next(self, c, segSub: bool = False):
        """
        If c == self, returns the first elemeht of the path

        If c is a pathElement, returns the following element of the path
        or None if there is no such element
        """
        if c is None:
            return None
        if c is self:
            if self.subpaths and self.subpaths[0]:
                t = self.subpaths[0][0]
                if not segSub or not isinstance(t.segment_sub, list):
                    return t
                else:
                    return t.segment_sub[0]
            return None
        assert ((segSub or not isinstance(c.segment_sub, int)) and
                (not segSub or not isinstance(c.segment_sub, list)))
        self.syncPositions()
        subpath, offset = c.position
        if isinstance(c.segment_sub, int):
            suboff = c.segment_sub + 1
            t = self.subpaths[subpath][offset]
            if len(t.segment_sub) > suboff:
                return t.segment_sub[suboff]
        offset += 1
        if offset < len(self.subpaths[subpath]):
            t = self.subpaths[subpath][offset]
            if not segSub or not isinstance(t.segment_sub, list):
                return t
            else:
                return t.segment_sub[0]
        subpath += 1
        offset = 0
        if subpath < len(self.subpaths):
            t = self.subpaths[subpath][offset]
            if not segSub or not isinstance(t.segment_sub, list):
                return t
            else:
                return t.segment_sub[0]
        return None

    # Iterate through path elements but start each subpath
    # with implicit closing line (if any). Skip pro forma
    # (zero-length) close elements
    def nextForHints(self, c):
        """
        Like next() but returns the next element in "hint order", with
        implicit close elements coming first in the subpath instead of
        last
        """
        if c is None:
            return None
        if c is self:
            if not self.subpaths or not self.subpaths[0]:
                return None
            c = self.subpaths[0][-1]
            if not c.isClose():
                c = self.subpaths[0][0]
            return c
        self.syncPositions()
        subpath, offset = c.position
        if c.isClose():
            n = self.subpaths[subpath][0]
            if n == c:
                return None
            return n
        offset += 1
        if offset < len(self.subpaths[subpath]) - 1:
            c = self.subpaths[subpath][offset]
            if not c.isClose():
                return c
        subpath += 1
        if subpath < len(self.subpaths):
            c = self.subpaths[subpath][-1]
            if not c.isClose():
                c = self.subpaths[subpath][0]
            return c
        return None

    def inSubpath(self, c, i, skipTiny, closeWrapOK, segSub):
        """Utility function for nextInSubpath and prevInSubpath"""
        if c is None or (c is self and i == -1):
            return None
        if c is self:
            return self.next(c, segSub)
        assert ((segSub or not isinstance(c.segment_sub, int)) and
                (not segSub or not isinstance(c.segment_sub, list)))
        self.syncPositions()
        subpath, offset = c.position
        sp = self.subpaths[subpath]
        l = len(sp)
        o = offset
        while True:
            done = False
            if isinstance(c.segment_sub, int):
                suboff = c.segment_sub + i
                t = self.subpaths[subpath][o]
                if 0 <= suboff < len(t.segment_sub):
                    c = t.segment_sub[suboff]
                    done = True
            if not done:
                o = (o + i) % l
                if o == offset:
                    return None
                if segSub and isinstance(sp[o].segment_sub, list):
                    if i < 0:
                        suboff = len(sp[o].segment_sub) + i
                    else:
                        suboff = 0
                    c = sp[o].segment_sub[suboff]
                else:
                    c = sp[o]
            if not skipTiny or not c.isTiny():
                break
            if (not closeWrapOK and
                    ((i > 0 and o < offset) or (i < 0 and o > offset))):
                return None
        return c

    def nextInSubpath(
        self,
        c,
        skipTiny: bool = False,
        closeWrapOK: bool = True,
        segSub: bool = False
    ):
        """
        Returns the next element in the subpath after c.

        If c is the last element and closeWrapOK is True returns the
        first element of the subpath.

        If c is the last element and closeWrapOK is False returns None
        """
        return self.inSubpath(c, 1, skipTiny, closeWrapOK, segSub)

    def prevInSubpath(
        self,
        c,
        skipTiny: bool = False,
        closeWrapOK: bool = True,
        segSub: bool = False
    ):
        """
        Returns the previous element in the subpath before c.

        If c is the first element and closeWrapOK is True returns the
        last element of the subpath.

        If c is the first element and closeWrapOK is False returns None
        """
        return self.inSubpath(c, -1, skipTiny, closeWrapOK, segSub)

    def nextSlopePoint(self, c):
        """Returns the slope point of the element of the subpath after c"""
        n = self.nextInSubpath(c, skipTiny=True, segSub=True)
        return None if n is None else n.slopePoint(0)

    def prevSlopePoint(self, c):
        """Returns the slope point of the element of the subpath before c"""
        p = self.prevInSubpath(c, skipTiny=True, segSub=True)
        return None if p is None else p.slopePoint(1)

    class glyphiter:
        """An iterator for a glyphData path"""

        __slots__ = ('gd', 'pos')

        def __init__(self, gd) -> None:
            self.gd = self.pos = gd

        def __next__(self):
            self.pos = self.gd.next(self.pos)
            if self.pos is None:
                raise StopIteration
            return self.pos

        def __iter__(self) -> Iterator:
            return self

    def __iter__(self) -> Iterator:
        return self.glyphiter(self)

    # utility

    def checkAssocPoint(self, segs, spe, ope, sp, op, mapEnd, loose,
                        factor=None) -> bool:
        if factor is None:
            factor = ope.getAssocFactor(loose)
        if sp == op or ope.containsPoint(sp, factor):
            if not ope.containsPoint(spe.atT(.5), factor * ope.middleMult):
                return False
            spe.association = ope
            return True
        else:
            does, t = spe.containsPoint(op, factor, True)
            if does and spe.tSlop < t < 1 - spe.tSlop:
                midMapped = (1 - (1 - t) / 2) if mapEnd else t / 2
                if not ope.containsPoint(spe.atT(midMapped),
                                         factor * ope.middleMult):
                    return False
                following = spe.splitAt(t)
                if mapEnd:
                    following.association = ope
                    sl = self.subpaths[spe.position[0]]
                    sl.insert(sl.index(spe) + 1, following)
                    segs.add(spe)
                else:
                    spe.association = ope
                    sl = self.subpaths[spe.position[0]]
                    sl.insert(sl.index(spe), following)
                    segs.add(following)
                self.setPathEdited()
                return True
        return False

    def associatePath(self, orig, loose: bool = False) -> None:
        peMap = defaultdict(list)
        for oc in orig:
            peMap[tuple(oc.e.round(1))].append(oc)
        segs = set()
        for s in self.subpaths:
            segs.update(s)
        # The position[0] subpath pointers for self pathElements should stay
        # accurate during the loop but the position[1] indexes won't.
        while segs:
            # Check a segment with the same start point
            c = segs.pop()
            done = False
            oepel = peMap.get(tuple(c.e.round(1)), [])
            if oepel:
                for oepe in oepel:
                    if self.checkAssocPoint(segs, c, oepe, c.s, oepe.s, True,
                                            loose):
                        done = True
                        break
                    else:
                        oepen = orig.nextInSubpath(oepe)
                        if self.checkAssocPoint(segs, c, oepen, c.s, oepen.e,
                                                True, loose):
                            done = True
                            break
                if done:
                    continue
            ospel = peMap.get(tuple(c.s.round(1)), [])
            if ospel:
                for ospe in ospel:
                    ospen = orig.nextInSubpath(ospe)
                    if self.checkAssocPoint(segs, c, ospen, c.e, ospen.e,
                                            False, loose):
                        done = True
                        break
                    elif self.checkAssocPoint(segs, c, ospe, c.e, ospe.s,
                                              False, loose):
                        done = True
                        break
                if done:
                    continue
            cBounds = c.getBounds()
            for oc in orig:
                factor = oc.getAssocFactor(loose)
                if cBounds.intersects(oc.getBounds(), factor):
                    if (oc.containsPoint(c.s, factor) and
                            (self.checkAssocPoint(segs, c, oc, c.e, oc.e,
                                                  False, loose, factor) or
                             self.checkAssocPoint(segs, c, oc, c.e, oc.s,
                                                  False, loose, factor))):
                        done = True
                        break
            if done:
                continue
            # Didn't find the start point anywhere, probably because it
            # was very close to but not quite identical to another start point.
            # Try again from the other end.
            for oc in orig:
                factor = oc.getAssocFactor(loose)
                if cBounds.intersects(oc.getBounds(), factor):
                    if (oc.containsPoint(c.e, factor) and
                            (self.checkAssocPoint(segs, c, oc, c.s, oc.s,
                                                  True, loose, factor) or
                             self.checkAssocPoint(segs, c, oc, c.s, oc.e,
                                                  True, loose, factor))):
                        done = True
                        break
            if done:
                continue
            assert not hasattr(c, 'association')
            log.warning("Unable to map derived path element from " +
                        "%s to %s" % (c.s, c.e))
            c.association = None
        self.syncPositions()

    def addNullClose(self, si) -> None:
        sp = self.subpaths[si]
        assert not sp[-1].isClose()
        pe = pathElement(sp[0].s, sp[0].s, is_close=True,
                         position=(si, len(sp)))
        sp.append(pe)

    def getStemMasks(self) -> list[Any] | None:
        """Utility function for pen methods"""
        s = self.nextmasks
        self.nextmasks = None
        return s

    def checkFlex(self, is_curve: bool):
        """Utility function for pen methods"""
        if self.nextflex is None:
            return None
        elif not is_curve:
            log.warning("Line marked with flex hint: ignoring")
            self.nextflex = None
            return None
        elif self.nextflex == 1:
            self.nextflex = 2
            return 1
        elif self.nextflex == 2:
            self.nextflex = None
            return 2
        else:
            log.error("Internal error: Bad flex value")
            self.nextflex = None
            return None

    def toStems(self, data):
        """Converts relative T2 charstring stem data to stem object array"""
        high = 0
        sl = []
        for i in range(len(data) // 2):
            low = high + data[i * 2]
            high = low + data[i * 2 + 1]
            sl.append(stem(low, high))  # pytype: disable=wrong-arg-count
        return sl

    def fromStems(self, stems):
        """Converts stem array to relative T2 charstring stem data"""
        l = None
        data = []
        for s in stems:
            data.extend(s.relVals(last=l))
            l = s
        return data

    def clearFlex(self) -> None:
        """Clears any flex hints"""
        if self.flex_count == 0:
            return
        for c in self:
            c.flex = None
        self.flex_count = 0

    def clearHints(self, doVert: bool = False) -> None:
        """Clears stem hints in specified dimension"""
        if doVert:
            self.vstems = []
            if self.startmasks and self.startmasks[0]:
                self.startmasks = [self.startmasks[0], None]
            else:
                self.startmasks = None
        else:
            self.hstems = []
            if self.startmasks and self.startmasks[1]:
                self.startmasks = [None, self.startmasks[1]]
            else:
                self.startmasks = None
        cntr = []
        for cn in self.cntr:
            if doVert and cn[0]:
                cntr.append((cn[0], None))
            elif not doVert and cn[1]:
                cntr.append((None, cn[1]))
        self.cntr = cntr
        for c in self:
            c.clearHints(doVert=doVert)

    def clearTempState(self) -> None:
        self.hhs = self.vhs = None
        for c in self:
            c.clearTempState()



# Copyright 2014 Adobe. All rights reserved.

from __future__ import annotations

"""
Utilities for converting between T2 charstrings and glyphData objects.
"""

import copy
import logging
import os
import subprocess
import tempfile
from fontTools.misc.psCharStrings import T2OutlineExtractor
from fontTools.ttLib import TTFont, newTable
from fontTools.misc.roundTools import noRound, otRound
from fontTools.varLib.varStore import VarStoreInstancer
from fontTools.varLib.cff import CFF2CharStringMergePen, MergeOutlineExtractor

# import subset.cff is needed to load the implementation for
# CFF.desubroutinize: the module adds this class method to the CFF and CFF2
# classes.
import fontTools.subset.cff
from typing import Any, Callable
from fontTools.cffLib import PrivateDict, TopDict
from fontTools.misc.psCharStrings import T2CharString
from fontTools.ttLib.tables._f_v_a_r import Axis
from fontTools.ttLib.tables.otTables import VarData

from . import fdTools, FontParseError
from .glyphData import glyphData

# keep linting tools quiet about unused import
assert fontTools.subset.cff is not None

log = logging.getLogger(__name__)

kStackLimit = 46
kStemLimit = 96


class SEACError(Exception):
    """Raised when a charString has an obsolete 'seac' operator"""

    pass


def _add_method(*clazzes) -> Callable:
    """Returns a decorator function that adds a new method to one or
    more classes."""

    def wrapper(method):
        done = []
        for clazz in clazzes:
            if clazz in done:
                continue  # Support multiple names of a clazz
            done.append(clazz)
            assert clazz.__name__ != 'DefaultTable', \
                'Oops, table class not found.'
            assert not hasattr(clazz, method.__name__), \
                "Oops, class '%s' has method '%s'." % (clazz.__name__,
                                                       method.__name__)
            setattr(clazz, method.__name__, method)
        return None
    return wrapper


def hintOn(i: int, hintMaskBytes: list) -> bool:
    """Used to help convert T2 hintmask bytes to a boolean array"""
    byteIndex = int(i / 8)
    byteValue = hintMaskBytes[byteIndex]
    offset = 7 - (i % 8)
    return ((2**offset) & byteValue) > 0


def denormalizeValue(nv: float, a: Axis) -> float:
    l, d, u = a.minValue, a.defaultValue, a.maxValue
    assert l <= d <= u
    if nv == 0:
        v = d
    elif nv < 0:
        v = l + (d - l) * (1 + nv)
    else:
        v = d + (u - d) * nv
    return v


def denormalizeLocation(
    loc: dict[str, float], axes: list[Axis]
) -> dict[str, float]:
    return {a.axisTag: denormalizeValue(loc.get(a.axisTag, a.defaultValue), a)
            for a in axes}


class T2ToGlyphDataExtractor(T2OutlineExtractor):
    """
    Inherits from the fontTools Outline Extractor and adapts some of the
    operator methods to match the "hint pen" interface of the glyphData class
    """

    def __init__(self, gd: glyphData, localSubrs, globalSubrs, nominalWidthX,
                 defaultWidthX, readStems=True, readFlex=True):
        T2OutlineExtractor.__init__(self, gd, localSubrs, globalSubrs,
                                    nominalWidthX, defaultWidthX)
        self.glyph = gd
        self.readStems = readStems
        self.readFlex = readFlex
        self.hintMaskBytes: int | None = None
        self.vhintCount = 0
        self.hhintCount = 0

    def op_endchar(self, index):
        self.endPath()
        args = self.popallWidth()
        if args:  # It is a 'seac' composite character. Don't process
            raise SEACError

    def op_hflex(self, index):
        if self.readFlex:
            self.glyph.nextIsFlex()
        super().op_hflex(index)

    def op_flex(self, index):
        if self.readFlex:
            self.glyph.nextIsFlex()
        super().op_flex(index)

    def op_hflex1(self, index):
        if self.readFlex:
            self.glyph.nextIsFlex()
        super().op_hflex1(index)

    def op_flex1(self, index):
        if self.readFlex:
            self.glyph.nextIsFlex()
        super().op_flex1(index)

    def op_hstem(self, index):
        args = self.popallWidth()
        self.countHints(args, True)
        if not self.readStems:
            return
        self.glyph.hStem(args, False)

    def op_vstem(self, index):
        args = self.popallWidth()
        self.countHints(args, False)
        if not self.readStems:
            return
        self.glyph.vStem(args, False)

    def op_hstemhm(self, index):
        args = self.popallWidth()
        self.countHints(args, True)
        if not self.readStems:
            return
        self.glyph.hStem(args, True)

    def op_vstemhm(self, index):
        args = self.popallWidth()
        self.countHints(args, False)
        if not self.readStems:
            return
        self.glyph.vStem(args, True)

    def countHints(self, args, horiz):
        if horiz:
            self.hhintCount += len(args) // 2
        else:
            self.vhintCount += len(args) // 2

    def doMask(self, index, cntr=False):
        args = []
        if not self.hintMaskBytes:
            args = self.popallWidth()
            if args:
                # values after stems operators followed by a
                # hint/cntrmask is an implicit vstem(hm))
                self.countHints(args, False)
                self.glyph.vStem(args, None if cntr else True)

            self.hintMaskBytes = (self.vhintCount + self.hhintCount + 7) // 8

        hintMaskString, index = self.callingStack[-1].getBytes(
            index, self.hintMaskBytes)

        tc = self.hhintCount + self.vhintCount
        hhints = [hintOn(i, hintMaskString) for i in range(self.hhintCount)]
        vhints = [hintOn(i, hintMaskString)
                  for i in range(self.hhintCount, tc)]
        return hintMaskString, hhints, vhints, index

    def op_hintmask(self, index):
        hintString, hhints, vhints, index = self.doMask(index)
        if self.readStems:
            self.glyph.hintmask(hhints, vhints)
        return hintString, index

    def op_cntrmask(self, index):
        hintString, hhints, vhints, index = self.doMask(index, True)
        if self.readStems:
            self.glyph.cntrmask(hhints, vhints)
        return hintString, index


def convertT2ToGlyphData(t2CharString, readStems=True, readFlex=True,
                         roundCoords=True, name=None):
    """Wrapper method for T2ToGlyphDataExtractor execution"""
    gl = glyphData(roundCoords=roundCoords, name=name)
    subrs = getattr(t2CharString.private, "Subrs", [])
    extractor = T2ToGlyphDataExtractor(gl, subrs,
                                       t2CharString.globalSubrs,
                                       t2CharString.private.nominalWidthX,
                                       t2CharString.private.defaultWidthX,
                                       readStems, readFlex)
    extractor.execute(t2CharString)
    t2_width_arg = None
    if extractor.gotWidth and (extractor.width is not None):
        t2_width_arg = extractor.width - t2CharString.private.nominalWidthX
    gl.setWidth(t2_width_arg)
    return gl


def _run_tx(args: list[str]) -> None:
    try:
        subprocess.check_call(["tx"] + args, stderr=subprocess.DEVNULL)
    except (subprocess.CalledProcessError, OSError) as e:
        raise FontParseError(e)


class CFFFontData:
    vsIndexMap: dict[str, tuple[int, bool]]

    def __init__(self, path: str, font_format: str) -> None:
        self.inputPath = path
        self.font_format = font_format
        self.is_cff2 = False
        self.is_vf = False
        self.vs_data_models: list[VarDataModel] | None = None
        self.desc = None
        if font_format == "OTF":
            # It is an OTF font, we can process it directly.
            font = TTFont(path)
            if "CFF " in font:
                cff_format = "CFF "
            elif "CFF2" in font:
                cff_format = "CFF2"
                self.is_cff2 = True
            else:
                raise FontParseError("OTF font has no CFF table <%s>." % path)
        else:
            # Else, package it in an OTF font.
            cff_format = "CFF "
            if font_format == "CFF":
                with open(path, "rb") as fp:
                    data = fp.read()
            else:
                fd, temp_path = tempfile.mkstemp()
                os.close(fd)
                try:
                    _run_tx(["-cff", "+b", "-std", path, temp_path])
                    with open(temp_path, "rb") as fp:
                        data = fp.read()
                finally:
                    os.remove(temp_path)

            font = TTFont()
            font['CFF '] = newTable('CFF ')
            font['CFF '].decompile(data, font)

        self.ttFont = font
        self.cffTable = font[cff_format]

        # for identifier in glyph-list:
        # Get charstring.
        self.topDict = self.cffTable.cff.topDictIndex[0]
        self.charStrings = self.topDict.CharStrings
        if 'fvar' in self.ttFont:
            # have not yet collected VF global data.
            self.is_vf = True
            self.glyph_programs: list[list[Any]] = []
            self.vsIndexMap = {}
            self.axes = self.ttFont['fvar'].axes
            CFF2 = self.cffTable
            CFF2.desubroutinize()
            topDict = CFF2.cff.topDictIndex[0]
            # We need a new charstring object into which we can save the
            # hinted CFF2 program data. Copying an existing charstring is a
            # little easier than creating a new one and making sure that all
            # properties are set correctly.
            self.temp_cs_extract = copy.deepcopy(self.charStrings['.notdef'])
            self.temp_cs_merge = copy.deepcopy(self.charStrings['.notdef'])
            self.vs_data_models = self.get_vs_data_models(topDict, self.axes)

    def getGlyphList(self) -> list[str]:
        return self.ttFont.getGlyphOrder()

    def getPSName(self) -> str:
        if self.is_cff2 and 'name' in self.ttFont:
            psName = next((name_rec.string for name_rec in self.ttFont[
                'name'].names if (name_rec.nameID == 6) and (
                    name_rec.platformID == 3)))
            psName = psName.decode('utf-16be')
        else:
            psName = self.cffTable.cff.fontNames[0]
        return psName

    def get_min_max(self, upm: int) -> tuple[int, int]:
        if self.is_cff2 and 'hhea' in self.ttFont:
            font_max = self.ttFont['hhea'].ascent
            font_min = self.ttFont['hhea'].descent
        elif hasattr(self.topDict, 'FontBBox'):
            font_max = self.topDict.FontBBox[3]
            font_min = self.topDict.FontBBox[1]
        else:
            font_max = int(upm * 1.25)
            font_min = int(-upm * 0.25)
        return font_min, font_max

    def convertToGlyphData(self, glyphName: str, readStems: bool,
                           readFlex: bool, roundCoords: bool,
                           doAll: bool = False) -> glyphData | None:
        t2CharString = self.charStrings[glyphName]
        try:
            gl = convertT2ToGlyphData(t2CharString, readStems, readFlex,
                                      roundCoords, name=glyphName)
        except SEACError:
            log.warning("Skipping %s: can't process SEAC composite glyphs.",
                        glyphName)
            gl = None
        return gl

    def updateFromGlyph(self, gl: glyphData | None, glyphName: str) -> None:
        if gl is None:
            return
        t2Program = gl.T2()

        if not self.is_cff2:
            t2_width_arg = gl.getWidth()
            if t2_width_arg is not None:
                t2Program = [t2_width_arg] + t2Program
        if self.vs_data_models is not None:
            # It is a variable font. Accumulate the charstrings.
            self.glyph_programs.append(t2Program)
        else:
            # This is an MM source font. Update the font's charstring directly.
            t2CharString = self.charStrings[glyphName]
            # Not used but useful for debugging save failures
            t2CharString._name = glyphName
            t2CharString.program = t2Program

    def save(self, path: str) -> None:
        if path is None:
            path = self.inputPath

        if self.font_format == "OTF":
            self.ttFont.save(path)
            self.ttFont.close()
        else:
            data = self.ttFont["CFF "].compile(self.ttFont)
            if self.font_format == "CFF":
                with open(path, "wb") as fp:
                    fp.write(data)
            else:
                fd, temp_path = tempfile.mkstemp()
                os.write(fd, data)
                os.close(fd)

                try:
                    args = ["-t1", "-std"]
                    if self.font_format == "PFB":
                        args.append("-pfb")
                    _run_tx(args + [temp_path, path])
                finally:
                    os.remove(temp_path)

    def close(self) -> None:
        self.ttFont.close()

    def isCID(self) -> bool:
        # XXX wrong for variable font
        return hasattr(self.topDict, "FDSelect")

    def getPrivateDictVal(
        self,
        pDict: PrivateDict,
        attr: str,
        default: int | list,
        vsindex: int | None,
        vsi: VarStoreInstancer | None
    ) -> int | float | list[float] | list[int]:
        val = getattr(pDict, attr, default)
        if attr in ('BlueValues', 'OtherBlues', 'FamilyBlues',
                    'FamilyOtherBlues', 'StemSnapH', 'StemSnapV'):
            last_def = 0
            last_calc = 0
            vlist: list[Any] = []
            for dl in val:  # type: ignore[union-attr]
                if isinstance(dl, list):
                    assert vsi is not None
                    calc = (dl[0] - last_def +
                            vsi.interpolateFromDeltas(vsindex, dl[1:]))
                    vlist.append(last_calc + calc)
                    last_def = dl[0]
                    last_calc += calc
                else:
                    vlist.append(dl)
                    last_def = last_calc = dl
            val = vlist
        else:
            if isinstance(val, list):
                assert vsi is not None
                val = val[0] + vsi.interpolateFromDeltas(vsindex, val[1:])

        return val

    def getPrivateFDDict(
        self,
        allowNoBlues: bool,
        noFlex: bool,
        vCounterGlyphs: dict[str, bool],
        hCounterGlyphs: dict[str, bool],
        desc: str,
        fdIndex: int = 0,
        gl_vsindex: int | None = None
    ) -> tuple[list[fdTools.FDDict], int] | fdTools.FDDict:
        pTopDict = self.topDict
        if hasattr(pTopDict, "FDArray"):
            pDict = pTopDict.FDArray[fdIndex]
        else:
            assert fdIndex == 0
            pDict = pTopDict
        privateDict = pDict.Private

        if self.is_vf:
            dict_vsindex = getattr(privateDict, "vsindex", 0)
            if gl_vsindex is None:
                gl_vsindex = dict_vsindex
            assert self.vs_data_models is not None
            vs_data_model = self.vs_data_models[gl_vsindex]
            vsi_list = vs_data_model.master_vsi_list
        else:
            assert gl_vsindex is None
            dict_vsindex = None
            vsi_list = [None]
            fontName = desc

        fdDictArray = []

        for vn, vsi in enumerate(vsi_list):
            if self.is_vf:
                assert gl_vsindex is not None  # Set above if was None
                fontName = self.getInstanceName(vn, vsi, self.axes)
                if fontName is None:
                    fontName = "Model %d Instance %d" % (gl_vsindex, vn)
            fdDict = fdTools.FDDict(fdIndex, fontName=fontName)
            fdDict.setInfo('LanguageGroup',
                           self.getPrivateDictVal(privateDict,
                                                  'LanguageGroup', 0,
                                                  dict_vsindex, vsi))

            if not self.is_cff2 and hasattr(pDict, "FontMatrix"):
                fontMatrix = pDict.FontMatrix
                upm = int(1 / fontMatrix[0])
            elif hasattr(pTopDict, 'FontMatrix'):
                fontMatrix = pTopDict.FontMatrix
                upm = int(1 / fontMatrix[0])
            else:
                upm = 1000
            fdDict.setInfo('OrigEmSqUnits', upm)

            for bvattr, bvkeys in (('BlueValues',
                                    fdTools.kBlueValueKeys),
                                   ('OtherBlues',
                                    fdTools.kOtherBlueValueKeys)):
                bvs = self.getPrivateDictVal(privateDict, bvattr, [],
                                             dict_vsindex, vsi)
                numbvs = len(bvs)  # type: ignore[arg-type]
                if bvattr == 'BlueValues' and numbvs < 4:
                    low, high = self.get_min_max(upm)
                    low = int(min(-upm * 0.25, low))
                    high = int(max(upm * 1.25, high))
                    # Make a set of inactive alignment zones: zones outside
                    # of the font BBox so as not to affect hinting. Used
                    # when source font has no BlueValues or has invalid
                    # BlueValues. Some fonts have bad BBox values, so I
                    # don't let this be smaller than -upm*0.25, upm*1.25.
                    inactiveAlignmentValues = [low, low, high, high]
                    if allowNoBlues or self.is_vf:
                        # XXX adjusted for vf
                        bvs = inactiveAlignmentValues
                        numbvs = len(bvs)
                    else:
                        raise FontParseError("Font must have at least four " +
                                             "values in its %s " % bvattr +
                                             "array for otfautohint to work!")
                    bvs.sort()  # type: ignore[union-attr]

                # The first pair only is a bottom zone, where the first value
                # is the overshoot position. The rest are top zones, and second
                # value of the pair is the overshoot position.
                if bvattr == 'BlueValues':
                    bvs[0] = bvs[0] - bvs[1]  # type: ignore[index, assignment]
                    for i in range(3, numbvs, 2):
                        bvs[i] = bvs[i] - bvs[i - 1]  # type: ignore[index, assignment]  # noqa: E501
                else:
                    for i in range(0, numbvs, 2):
                        bvs[i] = bvs[i] - bvs[i + 1]  # type: ignore[index, assignment]  # noqa: E501

                numbvs = min(numbvs, len(bvkeys))
                for i in range(numbvs):
                    key = bvkeys[i]
                    value = bvs[i]  # type: ignore[index]
                    fdDict.setInfo(key, value)

            for ssnap, stdw, fdkey in (('StemSnapV', 'StdVW', 'DominantV'),
                                       ('StemSnapH', 'StdHW', 'DominantH')):
                if hasattr(privateDict, ssnap):
                    sstems = self.getPrivateDictVal(privateDict, ssnap,
                                                    [], dict_vsindex, vsi)
                elif hasattr(privateDict, stdw):
                    sstems = [  # type: ignore[assignment]
                        self.getPrivateDictVal(
                            privateDict, stdw, -1, dict_vsindex, vsi)]
                else:
                    if allowNoBlues or self.is_vf:
                        # XXX adjusted for vf
                        # dummy value. Needs to be larger than any hint will
                        # likely be, as the autohint program strips out any
                        # hint wider than twice the largest global stem width.
                        sstems = [upm]
                    else:
                        raise FontParseError("Font has neither %s nor %s!" %
                                             (ssnap, stdw))
                sstems.sort()  # type: ignore[union-attr]
                if ((len(sstems) == 0) or  # type: ignore[arg-type]
                        ((len(sstems) == 1) and (sstems[0] < 1))):  # type: ignore[arg-type, index]  # noqa: E501
                    sstems = [upm]  # dummy value that will allow PyAC to run
                    log.warning("There is no value or 0 value for %s." % fdkey)
                fdDict.setInfo(fdkey, sstems)

            if noFlex:
                fdDict.setInfo('FlexOK', False)
            else:
                fdDict.setInfo('FlexOK', True)

            # Add candidate lists for counter hints, if any.
            if vCounterGlyphs:
                fdDict.setInfo('VCounterChars', vCounterGlyphs)
            if hCounterGlyphs:
                fdDict.setInfo('HCounterChars', hCounterGlyphs)

            fdDict.setInfo('BlueFuzz',
                           self.getPrivateDictVal(privateDict, 'BlueFuzz',
                                                  1, dict_vsindex, vsi))
            fdDict.buildBlueLists()

            if not self.is_vf:
                return fdDict

            fdDictArray.append(fdDict)

        assert self.is_vf and gl_vsindex is not None
        return fdDictArray, gl_vsindex

    def getfdIndex(self, name):
        try:
            gid = self.ttFont.getGlyphID(name)
        except KeyError:
            return None
        if gid is None:
            return None
        if hasattr(self.topDict, "FDSelect"):
            fdIndex = self.topDict.FDSelect[gid]
        else:
            fdIndex = 0
        return fdIndex

    def isVF(self) -> bool:
        return self.is_vf

    def getInputPath(self) -> str:
        return self.inputPath

    def getPrivateDict(self, fdIndex: int) -> PrivateDict:
        topDict = self.topDict
        if hasattr(topDict, "FDArray"):
            assert fdIndex < len(topDict.FDArray)
            private = topDict.FDArray[fdIndex].Private
        else:
            assert fdIndex == 0
            private = topDict.Private
        return private

    def mergePrivateMap(self, privateMap: dict[str, list[int]]) -> None:
        privateDict = self.getPrivateDict(0)
        for k, v in privateMap.items():
            setattr(privateDict, k, v)

    @staticmethod
    def getInstanceName(
        vn: int, vsi: VarStoreInstancer, axes: list[Axis]
    ) -> str:
        if vn == 0:
            vals = (f'{v:g}' for v in (a.defaultValue for a in axes))
            return 'DEF{' + ','.join(vals) + '}'
        else:
            vals = (f'{v:g}'
                    for v in denormalizeLocation(vsi.location, axes).values())
        return '{' + ','.join(vals) + '}'

    def get_vf_glyphs(self, glyph_name: str) -> tuple[list[glyphData], int]:
        charstring = self.charStrings[glyph_name]
        assert self.vs_data_models is not None

        if 'vsindex' in charstring.program:
            op_index = charstring.program.index('vsindex')
            vsindex = charstring.program[op_index - 1]
            self.vsIndexMap[glyph_name] = (vsindex, True)
        else:
            privateDict = self.getPrivateDict(self.getfdIndex(glyph_name))
            vsindex = getattr(privateDict, 'vsindex', 0)
            self.vsIndexMap[glyph_name] = (vsindex, False)
        vs_data_model = self.vs_data_models[vsindex]

        glyph_list = []
        for vsi in vs_data_model.master_vsi_list:
            iFD = vsi.interpolateFromDeltas
            t2_program = interpolate_cff2_charstring(charstring, glyph_name,
                                                     iFD, vsindex)
            self.temp_cs_extract.program = t2_program
            glyph = convertT2ToGlyphData(self.temp_cs_extract, True, False,
                                         True, name=glyph_name)
            glyph_list.append(glyph)

        return glyph_list, vsindex

    @staticmethod
    def get_vs_data_models(
        topDict: TopDict, axes: list[Axis]
    ) -> list[VarDataModel]:
        otvs = topDict.VarStore.otVarStore
        region_list = otvs.VarRegionList.Region
        axis_tags = [axis_entry.axisTag for axis_entry in axes]
        vs_data_models = []
        for vsindex, var_data in enumerate(otvs.VarData):
            vsi = VarStoreInstancer(topDict.VarStore.otVarStore, axes, {})
            master_vsi_list = [vsi]
            for region_idx in var_data.VarRegionIndex:
                region = region_list[region_idx]
                loc = {}
                for i, axis in enumerate(region.VarRegionAxis):
                    loc[axis_tags[i]] = axis.PeakCoord
                vsi = VarStoreInstancer(topDict.VarStore.otVarStore, axes,
                                        loc)
                master_vsi_list.append(vsi)
            vdm = VarDataModel(var_data, vsindex, master_vsi_list)
            vs_data_models.append(vdm)
        return vs_data_models

    def merge_hinted_glyphs(self, name: str) -> None:
        vsindex, addIndex = self.vsIndexMap[name]
        assert self.vs_data_models is not None
        new_t2cs = merge_hinted_programs(self.temp_cs_merge,
                                         self.glyph_programs, name,
                                         self.vs_data_models[vsindex])
        if addIndex:
            new_t2cs.program = [vsindex, 'vsindex'] + new_t2cs.program
        self.charStrings[name] = new_t2cs
        self.glyph_programs = []


def interpolate_cff2_charstring(charstring, gname, interpolateFromDeltas,
                                vsindex):
    # Interpolate charstring
    # e.g replace blend op args with regular args,
    # and discard vsindex op.
    new_program = []
    last_i = 0
    seenVOP = False
    program = charstring.program
    for i, token in enumerate(program):
        if token == 'vsindex':
            if last_i != 0:
                new_program.extend(program[last_i:i - 1])
            last_i = i + 1
            seenVOP = True
        elif token == 'blend':
            num_regions = charstring.getNumRegions(vsindex)
            numInstances = 1 + num_regions
            num_args = program[i - 1]
            # The program list starting at program[i] is now:
            # ..args for following operations
            # num_args values  from the default font
            # num_args tuples, each with numInstances-1 delta values
            # num_blend_args
            # 'blend'
            argi = i - (num_args * numInstances + 1)
            if last_i != argi:
                new_program.extend(program[last_i:argi])
            end_args = tuplei = argi + num_args
            master_args = []
            while argi < end_args:
                next_ti = tuplei + num_regions
                deltas = program[tuplei:next_ti]
                val = interpolateFromDeltas(vsindex, deltas)
                master_val = program[argi]
                master_val += otRound(val)
                master_args.append(master_val)
                tuplei = next_ti
                argi += 1
            new_program.extend(master_args)
            last_i = i + 1
            seenVOP = True
    if last_i != 0 or not seenVOP:
        new_program.extend(program[last_i:])
    return new_program


def merge_hinted_programs(
    charstring: T2CharString,
    t2_programs: list[list[str | bytes | int]],
    gname: str,
    vs_data_model: VarDataModel
) -> T2CharString:
    num_masters = vs_data_model.num_masters
    var_pen = CFF2CharStringMergePen([], gname, num_masters, 0)
    charstring.outlineExtractor = MergeOutlineExtractor

    for i, t2_program in enumerate(t2_programs):
        var_pen.restart(i)
        charstring.program = t2_program
        charstring.draw(var_pen)

    new_charstring = var_pen.getCharString(
        private=charstring.private,
        globalSubrs=charstring.globalSubrs,
        var_model=vs_data_model, optimize=True)
    return new_charstring


@_add_method(VarStoreInstancer)
def get_scalars(self, vsindex, region_idx) -> dict[int, float]:
    varData = self._varData
    # The index key needs to be the master value index, which includes
    # the default font value. VarRegionIndex provides the region indices.
    scalars = {0: 1.0}  # The default font always has a weight of 1.0
    region_index = varData[vsindex].VarRegionIndex
    for idx in range(region_idx):  # omit the scalar for the region.
        scalar = self._getScalar(region_index[idx])
        if scalar:
            scalars[idx + 1] = scalar
    return scalars


class VarDataModel(object):
    def __init__(
        self,
        var_data: VarData,
        vsindex: int,
        master_vsi_list: list[VarStoreInstancer]
    ) -> None:
        self.var_data = var_data
        self.master_vsi_list = master_vsi_list
        self._num_masters = len(master_vsi_list)

        # for default font value
        self.delta_weights: list[dict[int, float]] = [{}]
        for region_idx, vsi in enumerate(master_vsi_list[1:]):
            scalars = vsi.get_scalars(vsindex, region_idx)
            self.delta_weights.append(scalars)

    @property
    def num_masters(self) -> int:
        return self._num_masters

    def getDeltas(self, master_values, *, round=noRound) -> list[float]:
        assert len(master_values) == len(self.delta_weights)
        out: list[float] = []
        for i, scalars in enumerate(self.delta_weights):
            delta = master_values[i]
            for j, scalar in scalars.items():
                if scalar:
                    delta -= out[j] * scalar
            out.append(round(delta))
        return out



# Copyright 2014 Adobe. All rights reserved.

"""
Tools for processing "fontinfo" files. See "otfautohint --doc-fddict", or the
MakeOTF user guide for details on this format.

The "fontinfo" file is expected to be in the same directory as the input
font file.
"""

import logging
import re
import numbers
import sys
import os
from typing import Any, TYPE_CHECKING

from . import Number

if TYPE_CHECKING:
    from .__main__ import HintOptions, ReportOptions
    from .autohint import FontInstance
    from .otfFont import CFFFontData
    from .ufoFont import UFOFontData


log = logging.getLogger(__name__)

# Tokens seen in font info file that are not
# part of a FDDict or GlyphSet definition.
kBeginToken = "begin"
kEndToken = "end"
kFDDictToken = "FDDict"
kGlyphSetToken = "GlyphSet"
kFinalDictName = "FinalFont"
kDefaultDictName = "No Alignment Zones"
kBaseStateTokens = [
    "FontName",
    "FullName",
    "IsBoldStyle",
    "IsItalicStyle",
    "ConvertToCID",
    "PreferOS/2TypoMetrics",
    "IsOS/2WidthWeigthSlopeOnly",
    "IsOS/2OBLIQUE",
    "UseOldNameID4",
    "LicenseCode"
]
kBlueValueKeys = [
    "BaselineOvershoot",  # 0
    "BaselineYCoord",  # 1
    "CapHeight",  # 2
    "CapOvershoot",  # 3
    "LcHeight",  # 4
    "LcOvershoot",  # 5
    "AscenderHeight",  # 6
    "AscenderOvershoot",  # 7
    "FigHeight",  # 8
    "FigOvershoot",  # 9
    "Height5",  # 10
    "Height5Overshoot",  # 11
    "Height6",  # 12
    "Height6Overshoot",  # 13
]

kFamilyValueKeys = ['Family' + i for i in kBlueValueKeys]

kOtherBlueValueKeys = [
    "Baseline5Overshoot",  # 0
    "Baseline5",  # 1
    "Baseline6Overshoot",  # 2
    "Baseline6",  # 3
    "SuperiorOvershoot",  # 4
    "SuperiorBaseline",  # 5
    "OrdinalOvershoot",  # 6
    "OrdinalBaseline",  # 7
    "DescenderOvershoot",  # 8
    "DescenderHeight",  # 9
]

kOtherFamilyValueKeys = ['Family' + i for i in kOtherBlueValueKeys]

kOtherFDDictKeys = [
    "FontName",
    "OrigEmSqUnits",
    "LanguageGroup",
    "DominantV",
    "DominantH",
    "FlexOK",
    "VCounterChars",
    "HCounterChars",
    "BlueFuzz"
]

kFontDictBluePairsName = "BlueValuesPairs"
kFontDictOtherBluePairsName = "OtherBlueValuesPairs"
kFontDictFamilyPairsName = "FamilyValuesPairs"
kFontDictOtherFamilyPairsName = "OtherFamilyValuesPairs"

kRunTimeFDDictKeys = [
    "DictName",
    kFontDictBluePairsName,
    kFontDictOtherBluePairsName,
    kFontDictFamilyPairsName,
    kFontDictOtherFamilyPairsName,
]

kFDDictKeys = (kOtherFDDictKeys +
               kBlueValueKeys +
               kOtherBlueValueKeys +
               kFamilyValueKeys +
               kOtherFamilyValueKeys +
               kRunTimeFDDictKeys)
kFontInfoKeys = (kOtherFDDictKeys +
                 kBlueValueKeys +
                 kOtherBlueValueKeys +
                 kFamilyValueKeys +
                 kOtherFamilyValueKeys +
                 ["StemSnapH", "StemSnapV"])


BlueValue = tuple[int, int, str, str, int]
BlueZoneDict = dict[tuple[int, int], tuple[int, str, str]]


class FontInfoParseError(ValueError):
    pass


class FDDict:
    BlueValuesPairs: list[BlueValue]
    OtherBlueValuesPairs: list[BlueValue]
    BlueFuzz: int
    DominantH: list[int]
    DominantV: list[int]
    BaselineOvershoot: Number
    BaselineYCoord: Number
    CapHeight: Number
    CapOvershoot: Number
    FamilyBaselineYCoord: Number
    OrigEmSqUnits: Number
    FontName: str | None

    def __init__(
        self,
        fdIndex: int | None,
        dictName: str | None = None,
        fontName: str | None = None
    ) -> None:
        self.fdIndex = fdIndex
        for key in kFDDictKeys:
            setattr(self, key, None)
        if dictName is not None:
            self.DictName = dictName
        else:
            self.DictName = "FDArray[%s]" % fdIndex
        self.FontName = fontName
        self.FlexOK = True
        setattr(self, kFontDictBluePairsName, [])
        setattr(self, kFontDictOtherBluePairsName, [])
        setattr(self, kFontDictFamilyPairsName, [])
        setattr(self, kFontDictOtherFamilyPairsName, [])

    def __str__(self) -> str:
        a = ''
        for k, v in vars(self).items():
            if k not in kFontInfoKeys or v is None:
                continue
            if k == 'FontName':
                continue
            if isinstance(v, list):
                v = '[%s]' % ' '.join((str(i) for i in v))
            elif v is False:
                v = 'false'
            elif v is True:
                v = 'true'
            elif not isinstance(v, str):
                v = str(v)
            a += "%s %s\n" % (k, v)
        return a

    def setInfo(self, key, value) -> None:
        if key == 'LanguageGroup':
            if ((isinstance(value, numbers.Number) and value == 1) or
                    value == '1'):
                value = 1
            else:
                value = 0
        elif key in ('DominantV', 'DominantH'):
            if isinstance(value, list):
                value = [int(v) for v in value]
            else:
                value = [int(value)]
        elif key == 'FlexOK':
            if value is None:
                value = True        # default
            elif value == 'false':
                value = False
            else:
                value = bool(value)
        elif key in ('VCounterChars', 'HCounterChars'):
            pass  # already formatted in __main__::_parse_fontinfo_file
        elif key in ('FontName', 'DictName'):
            pass  # keep the string
        elif value is not None:
            value = int(value)

        setattr(self, key, value)

    def buildBlueLists(self):
        log.info("Building BlueValue lists for FDDict %s" % self.DictName)
        if self.BaselineOvershoot is None:
            raise FontInfoParseError(
                "FDDict definition %s is missing the BaselineYCoord/"
                "BaselineOvershoot values. These are required." %
                self.DictName)
        elif int(self.BaselineOvershoot) > 0:
            raise FontInfoParseError(
                "The BaselineYCoord/BaselineOvershoot in FDDict definition %s "
                "must be a bottom zone - the BaselineOvershoot must be "
                "negative, not positive." % self.DictName)

        blueKeyList = [kBlueValueKeys,
                       kOtherBlueValueKeys,
                       kFamilyValueKeys,
                       kOtherFamilyValueKeys]
        bluePairListNames = [kFontDictBluePairsName,
                             kFontDictOtherBluePairsName,
                             kFontDictFamilyPairsName,
                             kFontDictOtherFamilyPairsName]
        for i in range(4):
            keyList = blueKeyList[i]
            pairFieldName = bluePairListNames[i]
            bluePairList = []
            for key in keyList:
                if key.endswith("Overshoot"):
                    width = getattr(self, key)
                    if width is not None:
                        baseName = key[:-len("Overshoot")]
                        zonePos = None
                        if key == "BaselineOvershoot":
                            zonePos = self.BaselineYCoord
                            tempKey = "BaselineYCoord"
                        elif key == "FamilyBaselineOvershoot":
                            zonePos = self.FamilyBaselineYCoord
                            tempKey = "FamilyBaselineYCoord"
                        else:
                            for posSuffix in ["", "Height", "Baseline"]:
                                tempKey = "%s%s" % (baseName, posSuffix)
                                value = getattr(self, tempKey, None)
                                if value is not None:
                                    zonePos = value
                                    break
                        if zonePos is None:
                            raise FontInfoParseError(
                                "Failed to find fontinfo FDDict %s top/bottom "
                                "zone name %s to match the zone width key '%s'"
                                "." % (self.DictName, tempKey, key))
                        if width < 0:
                            topPos = zonePos
                            bottomPos = zonePos + width
                            isBottomZone = 1
                            if (i == 0) and (key != "BaselineOvershoot"):
                                raise FontInfoParseError(
                                    "FontDict %s. Zone %s is a top zone, and "
                                    "the width (%s) must be positive." %
                                    (self.DictName, tempKey, width))
                            elif ((i == 2) and
                                  (key != "FamilyBaselineOvershoot")):
                                raise FontInfoParseError(
                                    "FontDict %s. Zone %s is a top zone, and "
                                    "the width (%s) must be positive." %
                                    (self.DictName, tempKey, width))
                        else:
                            bottomPos = zonePos
                            topPos = zonePos + width
                            isBottomZone = 0
                            if i == 1 and width > 0:
                                raise FontInfoParseError(
                                    "FontDict %s. Zone %s is a bottom zone, "
                                    "and so the width (%s) must be negative." %
                                    (self.DictName, tempKey, width))
                        bluePairList.append((topPos, bottomPos, tempKey,
                                            self.DictName, isBottomZone))
                        log.debug("%s BlueValue %s: (%g, %g)" %
                                  ('Bottom' if isBottomZone else 'Top',
                                   tempKey, bottomPos, topPos))
            if bluePairList:
                bluePairList = sorted(bluePairList)
                prevPair = bluePairList[0]
                zoneBuffer = 2 * self.BlueFuzz + 1
                for pair in bluePairList[1:]:
                    if prevPair[0] > pair[1]:
                        raise FontInfoParseError(
                            "In FDDict %s. The top of zone %s at %s overlaps "
                            "zone %s with the bottom at %s." %
                            (self.DictName, prevPair[2], prevPair[0], pair[2],
                             pair[1]))
                    elif abs(pair[1] - prevPair[0]) < zoneBuffer:
                        raise FontInfoParseError(
                            "In FDDict %s. The top of zone %s at %s is within "
                            "the min separation limit (%s units) of zone %s "
                            "with the bottom at %s." %
                            (self.DictName, prevPair[2], prevPair[0],
                             zoneBuffer, pair[2], pair[1]))
                    prevPair = pair
                setattr(self, pairFieldName, bluePairList)
        return

    def __repr__(self) -> str:
        fddict = {k: v for k, v in vars(self).items() if v is not None}
        return "<%s '%s' %s>" % (
            self.__class__.__name__, fddict.get('DictName', 'no name'), fddict)


def parseFontInfoFile(
    fdArrayMap: dict[int, FDDict],
    data: str,
    glyphList: list[str],
    maxY: int,
    minY: int,
    fontName: str
) -> tuple[dict[str, int], FDDict | None]:
    # fontDictList may or may not already contain a
    # font dict taken from the source font top FontDict.
    # The map of glyph names to font dict: the index into fontDictList.
    fdSelectMap = {}
    # The user-specified set of blue values to write into the output font,
    # some sort of merge of the individual font dicts. May not be supplied.
    finalFDict = None
    setName = None

    glyphSetSet = set()

    blueFuzz = fdArrayMap[0].BlueFuzz

    # Get rid of comments.
    data = re.sub(r"#[^\r\n]+[\r\n]", "", data)

    # We assume that no items contain whitespace.
    tokenList = data.split()
    numTokens = len(tokenList)
    i = 0
    baseState = 0
    inValue = 1
    inDictValue = 2
    dictState = 3
    glyphSetState = 4
    fdIndexDict = {}
    assert tuple(fdArrayMap.keys()) == (0,)
    maxfdIndex = max(fdArrayMap.keys())
    dictValueList = []
    dictKeyWord = ''
    fdDict: FDDict | None = None

    state = baseState

    while i < numTokens:
        token = tokenList[i]
        i += 1
        if state == baseState:
            if token == kBeginToken:
                token = tokenList[i]
                i += 1
                if token == kFDDictToken:
                    state = dictState
                    dictName = tokenList[i]
                    i += 1
                    if dictName == kFinalDictName:
                        # This dict is NOT used to hint any glyphs; it is
                        # used to supply the merged alignment zones and stem
                        # widths for the final font.
                        if finalFDict is None:
                            finalFDict = FDDict(maxfdIndex + 1,
                                                dictName=dictName)
                        fdDict = finalFDict
                    else:
                        if dictName not in fdIndexDict:
                            fdDict = FDDict(maxfdIndex + 1, dictName=dictName)
                            # save dict and FDIndex.
                            maxfdIndex += 1
                            fdIndexDict[dictName] = maxfdIndex
                            fdArrayMap[maxfdIndex] = fdDict
                        else:
                            fdDict = fdArrayMap[fdIndexDict[dictName]]

                elif token == kGlyphSetToken:
                    state = glyphSetState
                    setName = tokenList[i]
                    if setName in glyphSetSet:
                        raise FontInfoParseError(
                            "Duplicate GlyphSet \"%s\"" % setName)
                    else:
                        glyphSetSet.add(setName)
                    i += 1
                else:
                    raise FontInfoParseError(
                        "Unrecognized token after \"begin\" keyword: %s" %
                        token)

            elif token in kBaseStateTokens:
                # Discard value for base token.
                token = tokenList[i]
                i += 1
                if (token[0] in ["[", "("]) and (not token[-1] in ["]", ")"]):
                    state = inValue
            else:
                raise FontInfoParseError(
                    "Unrecognized token in base state: %s" % token)

        elif state == inValue:
            # We are processing a list value for a base state token.
            if token[-1] in ["]", ")"]:
                state = baseState  # found the last token in the list value.

        elif state == inDictValue:
            if token[-1] in ["]", ")"]:
                dictValueList.append(token[:-1])
                assert fdDict is not None
                fdDict.setInfo(dictKeyWord, dictValueList)
                state = dictState  # found the last token in the list value.
            else:
                dictValueList.append(token)

        elif state == glyphSetState:
            # "end GlyphSet" marks end of set,
            # else we are adding a new glyph name.
            if (token == kEndToken) and tokenList[i] == kGlyphSetToken:
                if tokenList[i + 1] != setName:
                    raise FontInfoParseError(
                        "End glyph set name \"%s\" does not match begin glyph "
                        "set name \"%s\"." % (tokenList[i + 1], setName))
                state = baseState
                i += 2
                setName = None
            else:
                # Need to add matching glyphs.
                gi = 0
                for gname in glyphList:
                    if re.search(token, gname):
                        # fdIndex value
                        assert setName is not None
                        fdSelectMap[gname] = fdIndexDict[setName]
                    gi += 1

        elif state == dictState:
            # "end FDDict" marks end of set,
            # else we are adding a new glyph name.
            if (token == kEndToken) and tokenList[i] == kFDDictToken:
                if tokenList[i + 1] != dictName:
                    raise FontInfoParseError(
                        "End FDDict  name \"%s\" does not match begin FDDict "
                        "name \"%s\"." % (tokenList[i + 1], dictName))
                state = baseState
                i += 2
                dictName = None
                fdDict = None
            else:
                assert fdDict is not None
                if token in kFDDictKeys:
                    value = tokenList[i]
                    i += 1
                    if value[0] in ["[", "("]:
                        if not value[-1] in ["]", ")"]:
                            state = inDictValue
                            dictValueList = [value[1:]]
                            dictKeyWord = token
                        else:
                            fdDict.setInfo(token, value[1:-1])
                    else:
                        fdDict.setInfo(token, value)
                else:
                    raise FontInfoParseError(
                        "FDDict key \"%s\" in fdDict named \"%s\" is not "
                        "recognized." % (token, dictName))

    for dictName, fdIndex in fdIndexDict.items():
        fdDict = fdArrayMap[fdIndex]
        assert fdDict is not None
        if fdDict.DominantH is None:
            log.warning("The FDDict '%s' in fontinfo has no "
                        "DominantH value", dictName)
        if fdDict.DominantV is None:
            log.warning("The FDDict '%s' in fontinfo has no "
                        "DominantV value", dictName)
        if fdDict.BlueFuzz is None:
            fdDict.BlueFuzz = blueFuzz
        fdDict.buildBlueLists()
        if fdDict.FontName is None:
            fdDict.FontName = fontName

    if fdIndexDict:
        # There are some FDDict definitions. This means that we need
        # to fix the default fontDict, inherited from the source font,
        # so that it has blues zones that will not affect hinting,
        # e.g outside of the Font BBox. We do this becuase if there are
        # glyphs which are not assigned toa user specified font dict,
        # it is becuase it doesn't make sense to provide alignment zones
        # for the glyph. Since otfautohint does require at least one bottom
        # zone and one top zone, we add one bottom and one top zone that are
        # outside the font BBox, so that hinting won't be affected by them.
        for gname in glyphList:
            if gname not in fdSelectMap:
                fdSelectMap[gname] = 0
        defaultFDDict = fdArrayMap[0]
        for key in kBlueValueKeys + kOtherBlueValueKeys:
            defaultFDDict.setInfo(key, None)
        defaultFDDict.BaselineYCoord = minY - 100
        defaultFDDict.BaselineOvershoot = 0  # type: ignore[attr-defined]
        defaultFDDict.CapHeight = maxY + 100  # type: ignore[attr-defined]
        defaultFDDict.CapOvershoot = 0  # type: ignore[attr-defined]
        defaultFDDict.BlueFuzz = 0
        defaultFDDict.DictName = kDefaultDictName  # "No Alignment Zones"
        defaultFDDict.FontName = fontName
        defaultFDDict.buildBlueLists()

    return fdSelectMap, finalFDict


def mergeFDDicts(prevDictList: list[FDDict]) -> dict[str, Any]:
    # Extract the union of the stem widths and zones from the list
    # of FDDicts, and replace the current values in the topDict.
    blueZoneDict: BlueZoneDict = {}
    otherBlueZoneDict: BlueZoneDict = {}
    dominantHDict: dict[int, str] = {}
    dominantVDict: dict[int, str] = {}
    bluePairListNames = [kFontDictBluePairsName, kFontDictOtherBluePairsName]
    zoneDictList = [blueZoneDict, otherBlueZoneDict]
    for prefDDict in prevDictList:
        if prefDDict is None:
            continue
        for i in (0, 1):
            zoneDict = zoneDictList[i]
            bluePairName = bluePairListNames[i]
            bluePairList = getattr(prefDDict, bluePairName)
            if not bluePairList:
                continue
            for topPos, bottomPos, zoneName, _, isBottomZone in bluePairList:
                zoneDict[(topPos, bottomPos)] = (isBottomZone, zoneName,
                                                 prefDDict.DictName)
        # Now for the stem widths.
        stemNameList = ["DominantH", "DominantV"]
        stemDictList = [dominantHDict, dominantVDict]
        for i in (0, 1):
            stemFieldName = stemNameList[i]
            dList: list[int] = getattr(prefDDict, stemFieldName)
            stemDict = stemDictList[i]
            if dList is not None:
                for width in dList:
                    stemDict[width] = prefDDict.DictName
    assert prefDDict
    # Now we have collected all the stem widths and zones
    # from all the dicts. See if we can merge them.
    goodBlueZoneList: list[int] = []
    goodOtherBlueZoneList: list[int] = []
    goodHStemList: list[int] = []
    goodVStemList: list[int] = []

    zoneDictList = [blueZoneDict, otherBlueZoneDict]
    goodZoneLists = [goodBlueZoneList, goodOtherBlueZoneList]
    stemDictList = [dominantHDict, dominantVDict]
    goodStemLists = [goodHStemList, goodVStemList]

    for ki in [0, 1]:
        zoneDict = zoneDictList[ki]
        goodZoneList = goodZoneLists[ki]
        stemDict = stemDictList[ki]
        goodStemList = goodStemLists[ki]

        # Zones first.
        zoneList = sorted(zoneDict.keys())
        if not zoneList:
            continue
        # Now check for conflicts.
        prevZone = zoneList[0]
        goodZoneList.append(prevZone[1])
        goodZoneList.append(prevZone[0])
        zoneBuffer = 2 * prefDDict.BlueFuzz + 1
        for zone in zoneList[1:]:
            curEntry = zoneDict[zone]
            prevEntry = zoneDict[prevZone]
            zoneName = curEntry[1]
            fdDictName = curEntry[2]
            prevZoneName = prevEntry[1]
            prevFDictName = prevEntry[2]

            if zone[1] < prevZone[0]:
                log.warning("For final FontDict, skipping zone %s in FDDict %s"
                            " because it overlaps with zone %s in FDDict %s.",
                            zoneName, fdDictName, prevZoneName, prevFDictName)
            elif abs(zone[1] - prevZone[0]) < zoneBuffer:
                log.warning("For final FontDict, skipping zone %s in FDDict %s"
                            " because it is within the minimum separation "
                            "allowed (%s units) of %s in FDDict %s.",
                            zoneName, fdDictName, zoneBuffer, prevZoneName,
                            prevFDictName)
            else:
                goodZoneList.append(zone[1])
                goodZoneList.append(zone[0])

            prevZone = zone

        stemList = sorted(stemDict.keys())
        if not stemList:
            continue
        # Now check for conflicts.
        prevStem = stemList[0]
        goodStemList.append(prevStem)
        for stem in stemList[1:]:
            if abs(stem - prevStem) < 2:
                log.warning("For final FontDict, skipping stem width %s in "
                            "FDDict %s because it overlaps in coverage with "
                            "stem width %s in FDDict %s.",
                            stem, stemDict[stem], prevStem, stemDict[prevStem])
            else:
                goodStemList.append(stem)
            prevStem = stem
    privateMap: dict[str, Any] = {}
    if goodBlueZoneList:
        privateMap['BlueValues'] = goodBlueZoneList
    if goodOtherBlueZoneList:
        privateMap['OtherBlues'] = goodOtherBlueZoneList
    else:
        privateMap['OtherBlues'] = None
    if goodHStemList:
        privateMap['StemSnapH'] = goodHStemList
    else:
        privateMap['StemSnapH'] = None
    if goodVStemList:
        privateMap['StemSnapV'] = goodVStemList
    else:
        privateMap['StemSnapV'] = None
    return privateMap


def fontinfoIncludeData(fdir, idir, match):
    incrpath = match.group(1)
    incpath = os.path.join(fdir, incrpath)
    if os.path.isfile(incpath):
        with open(incpath, "r", encoding="utf-8") as ii:
            return ii.read()
    if idir is not None:
        incpath = os.path.join(idir, incrpath)
        if os.path.isfile(incpath):
            with open(incpath, "r", encoding="utf-8") as ii:
                return ii.read()
    raise FontInfoParseError("fontinfo include file \"%s\" not found" %
                             incrpath)


def fontinfoFileData(
    options: "HintOptions | ReportOptions",
    font: "UFOFontData | CFFFontData"
) -> tuple[str, bool] | tuple[None, None]:
    # Don't use fontinfo file for a variable font
    if not font.isVF():
        fdir = os.path.dirname(font.getInputPath())
        if options.fontinfoPath is not None:
            idir = os.path.dirname(options.fontinfoPath)
            srcFontInfo = options.fontinfoPath
        else:
            idir = None
            srcFontInfo = os.path.join(fdir, "fontinfo")
        if os.path.isfile(srcFontInfo):
            with open(srcFontInfo, "r", encoding="utf-8") as fi:
                fontInfoData = fi.read()
            # Process includes
            subsMade = 1
            while subsMade > 0:
                fontInfoData, subsMade = re.subn(
                    r"^include (.*)$",
                    lambda m: fontinfoIncludeData(fdir, idir, m),
                    fontInfoData, count=0, flags=re.M)
            fontInfoData = re.sub(r"#[^\r\n]+", "", fontInfoData)
            return fontInfoData, 'FDDict' in fontInfoData
    return None, None


def getFDInfo(
    font: "UFOFontData | CFFFontData",
    desc: str,
    options: "HintOptions | ReportOptions",
    glyphList: list[str],
    isVF: bool
) -> tuple[dict[str, int], Any, Any]:
    privateMap = None
    # Check the fontinfo file, and add any other font dicts
    srcFontinfoData, hasDict = fontinfoFileData(options, font)
    if hasDict and not options.ignoreFontinfo:
        # Get the default fontinfo from the font's top dict.
        fdDict = font.getPrivateFDDict(options.allowNoBlues,
                                       options.noFlex,
                                       options.vCounterGlyphs,
                                       options.hCounterGlyphs, desc)
        assert isinstance(fdDict, FDDict)
        fdArrayMap: dict[int, FDDict] = {0: fdDict}
        minY, maxY = font.get_min_max(int(fdDict.OrigEmSqUnits))
        assert srcFontinfoData is not None
        fdSelectMap, finalFDict = parseFontInfoFile(
            fdArrayMap, srcFontinfoData, glyphList, int(maxY), int(minY), desc)
        if finalFDict is None:
            # If a font dict was not explicitly specified for the
            # output font, use the first user-specified font dict.
            fontDictList = [v for k, v in sorted(fdArrayMap.items())]
            privateMap = mergeFDDicts(fontDictList[1:])
        else:
            privateMap = mergeFDDicts([finalFDict])
    elif isVF:
        fdSelectMap = {}
        dictRecord: dict[int, dict[int, FDDict]] = {}
        fdArraySet = set()
        for name in glyphList:
            fdIndex = font.getfdIndex(name)
            if fdIndex is None:
                continue
            if fdIndex not in fdArraySet:
                result = font.getPrivateFDDict(options.allowNoBlues,
                                               options.noFlex,
                                               options.vCounterGlyphs,
                                               options.hCounterGlyphs,
                                               desc, fdIndex)
                assert isinstance(result, tuple)
                fddict, model = result
                fdArraySet.add(fdIndex)
                if model not in dictRecord:
                    dictRecord[model] = {}
                dictRecord[model][fdIndex] = fddict  # type: ignore
            fdSelectMap[name] = fdIndex
        return fdSelectMap, dictRecord, None
    else:
        fdSelectMap = {}
        fdArrayMap = {}
        for name in glyphList:
            fdIndex = font.getfdIndex(name)
            if fdIndex is None:
                continue
            if fdIndex not in fdArrayMap:
                fddict = font.getPrivateFDDict(  # type: ignore[assignment]
                    options.allowNoBlues, options.noFlex,
                    options.vCounterGlyphs, options.hCounterGlyphs,
                    desc, fdIndex)
                assert isinstance(fddict, FDDict)
                fdArrayMap[fdIndex] = fddict
            fdSelectMap[name] = fdIndex

    return fdSelectMap, fdArrayMap, privateMap


class FDDictManager:
    def __init__(
        self,
        options: "HintOptions | ReportOptions",
        fontInstances: list["FontInstance"],
        glyphList: list[str],
        isVF: bool = False
    ) -> None:
        self.options = options
        self.fontInstances = fontInstances
        self.glyphList = glyphList
        self.isVF = isVF
        self.fdSelectMap: dict | None = None
        self.auxRecord: dict[int, Any] = {}
        refI = fontInstances[0]
        fdArrayCompat = True
        fdArrays = []
        assert not isVF or len(fontInstances) == 1

        for i in fontInstances:
            (fdSelectMap, fdArrayMap,
             privateMap) = getFDInfo(i.font, i.font.desc, options,
                                     glyphList, isVF)
            if isVF:
                self.fdSelectMap = fdSelectMap
                self.dictRecord = fdArrayMap
            else:
                fdArrays.append(fdArrayMap)
                if self.fdSelectMap is None:
                    self.fdSelectMap = fdSelectMap
                elif not self.addDict(self.fdSelectMap, fdSelectMap):
                    log.error("FDDict indexes for font %s " % i.font.desc +
                              ("different from those in font %s" %
                               refI.font.desc))
                    fdArrayCompat = False
                if privateMap is not None:
                    # XXX make this decision more selective
                    i.font.mergePrivateMap(privateMap)

        if not isVF:
            if fdArrayCompat:
                fdmap = {}
                for fdIndex in fdArrays[0].keys():
                    fdmap[fdIndex] = [x[fdIndex] if fdIndex in x else None
                                      for x in fdArrays]
                self.dictRecord = {0: fdmap}  # type: ignore[assignment]
            else:
                log.error("Cannot continue")
                sys.exit()

        assert self.fdSelectMap
        if options.printFDDictList or options.printAllFDDict:
            # Print the user defined FontDicts, and exit.
            print("Private Dictionaries:\n")
            for _model, fda in self.dictRecord.items():
                for fdIndex, fdl in fda.items():
                    for fontDict in fdl:
                        if fontDict is None:
                            continue
                        print(fontDict.FontName)
                        print(fontDict.DictName)
                        print(str(fontDict))
                        gnameList = [gn for gn, fi in self.fdSelectMap.items()
                                     if fdIndex == fi]
                        print("%d glyphs:" % len(gnameList))
                        if len(gnameList) > 0:
                            gTxt = " ".join(gnameList)
                        else:
                            gTxt = "None"
                        print(gTxt + "\n")
                        if not options.printAllFDDict:
                            break
        elif len(set(self.fdSelectMap.keys())) > 1:
            log.info("Using different FDDicts for some glyphs.")

        self.checkGlyphList()

    def getDictRecord(self):
        return self.dictRecord

    def getRecKey(self, gname: str, vsindex):
        assert self.fdSelectMap is not None
        fdIndex = self.fdSelectMap[gname]
        if vsindex in self.dictRecord and fdIndex in self.dictRecord[vsindex]:
            return vsindex, fdIndex
        assert self.isVF
        if vsindex not in self.auxRecord:
            self.auxRecord[vsindex] = {}
        elif fdIndex in self.auxRecord[vsindex]:
            return self.auxRecord[vsindex][fdIndex]
        f = self.fontInstances[0]
        fddict, _ = f.font.getPrivateFDDict(self.options.allowNoBlues,
                                            self.options.noFlex,
                                            self.options.vCounterGlyphs,
                                            self.options.hCounterGlyphs,
                                            f.font.desc, fdIndex, vsindex)
        self.auxRecord[vsindex][fdIndex] = fddict
        return fddict

    def checkGlyphList(self) -> None:
        options = self.options
        glyphSet = set(self.glyphList)
        # Check for missing glyphs explicitly added via fontinfo or cmd line
        for label, charDict in [("hCounterGlyphs", options.hCounterGlyphs),
                                ("vCounterGlyphs", options.vCounterGlyphs),
                                ("upperSpecials", options.upperSpecials),
                                ("lowerSpecials", options.lowerSpecials),
                                ("noBlues", options.noBlues)]:
            for name in (n for n, w in charDict.items()
                         if w and n not in glyphSet):
                log.warning("%s glyph named in fontinfo is " % label +
                            "not in font: %s" % name)

    def addDict(self, dict1: dict, dict2: dict) -> bool:
        # This allows for sparse masters, just verifying that if a glyph name
        # is in both dictionaries it maps to the same index.
        good = True
        for k, v in dict2.items():
            if k not in dict1:
                dict1[k] = v
            elif v != dict1[k]:
                good = False
        return good



"""
Shell completion script generator for afdko command.

Generates completion scripts for bash, zsh, fish, and PowerShell.
"""

import sys
from typing import List, Tuple


SUPPORTED_SHELLS = ['bash', 'zsh', 'fish', 'powershell']


def get_commands() -> List[Tuple[str, str]]:
    """Extract full command names and descriptions from invoker.

    Returns list of (name, description) tuples, excluding abbreviations.
    """
    from afdko.invoker import ALL_COMMANDS
    commands = []
    seen_targets = set()

    for name, (target, desc, _) in ALL_COMMANDS.items():
        # Skip abbreviations (they have desc=None)
        # Skip duplicate targets (same function, different name)
        if desc is not None and target not in seen_targets:
            commands.append((name, desc))
            seen_targets.add(target)

    return sorted(commands)


def generate_bash(commands: List[Tuple[str, str]]) -> str:
    """Generate bash completion script."""
    cmd_list = ' '.join(name for name, _ in commands)

    return f'''# bash completion for afdko
# Save this file or add to your ~/.bashrc:
#   eval "$(afdko completion bash)"

_afdko_completion() {{
    local cur="${{COMP_WORDS[COMP_CWORD]}}"
    local prev="${{COMP_WORDS[COMP_CWORD-1]}}"

    # Only complete the first argument (the command)
    if [ "$COMP_CWORD" -eq 1 ]; then
        local commands="{cmd_list}"
        COMPREPLY=($(compgen -W "$commands" -- "$cur"))
    fi
}}

complete -F _afdko_completion afdko
'''


def generate_zsh(commands: List[Tuple[str, str]]) -> str:
    """Generate zsh completion script."""
    # Build command list with descriptions
    cmd_lines = '\n        '.join(
        f"'{name}:{desc}'" for name, desc in commands
    )

    return f'''#compdef afdko
# zsh completion for afdko
# Save to a directory in your $fpath, e.g.:
#   afdko completion zsh > /usr/local/share/zsh/site-functions/_afdko
# Or add to ~/.zshrc:
#   eval "$(afdko completion zsh)"

_afdko() {{
    local -a commands
    commands=(
        {cmd_lines}
    )

    _arguments '1: :->command' '*::arg:->args'

    case $state in
        command)
            _describe 'afdko command' commands
            ;;
    esac
}}

_afdko "$@"
'''


def generate_fish(commands: List[Tuple[str, str]]) -> str:
    """Generate fish completion script."""
    # Generate one complete line per command
    complete_lines = '\n'.join(
        f'complete -c afdko -f -n "__fish_use_subcommand" '
        f'-a {name} -d "{desc}"'
        for name, desc in commands
    )

    return f'''# fish completion for afdko
# Save to ~/.config/fish/completions/afdko.fish:
#   afdko completion fish > ~/.config/fish/completions/afdko.fish

{complete_lines}
'''


def generate_powershell(commands: List[Tuple[str, str]]) -> str:
    """Generate PowerShell completion script."""
    # Build command array
    cmd_list = ', '.join(f"'{name}'" for name, _ in commands)

    return rf'''# PowerShell completion for afdko
# Add to your PowerShell profile:
#   afdko completion powershell | Out-String | Invoke-Expression
# Or open profile with:
#   notepad $profile

Register-ArgumentCompleter -Native -CommandName afdko -ScriptBlock {{
    param($wordToComplete, $commandAst, $cursorPosition)

    # Get the current line and cursor position
    $line = $commandAst.ToString()

    # Only complete if we're completing the first argument
    $words = $line -split '\s+'
    if ($words.Count -le 2) {{
        $commands = @({cmd_list})

        $commands | Where-Object {{ $_ -like "$wordToComplete*" }} |
            ForEach-Object {{
            [System.Management.Automation.CompletionResult]::new(
                $_,
                $_,
                'ParameterValue',
                $_
            )
        }}
    }}
}}
'''


def print_help() -> None:
    """Print help message."""
    print("Usage: afdko completion <shell>")
    print()
    print("Generate shell completion script for afdko commands.")
    print()
    print("Supported shells:")
    print("  bash       Bourne Again Shell")
    print("  zsh        Z Shell")
    print("  fish       Friendly Interactive Shell")
    print("  powershell PowerShell")
    print()
    print("Examples:")
    print("  # bash - add to ~/.bashrc or ~/.bash_profile")
    print('  eval "$(afdko completion bash)"')
    print()
    print("  # zsh - save to a directory in $fpath")
    print("  afdko completion zsh > "
          "/usr/local/share/zsh/site-functions/_afdko")
    print()
    print("  # fish - save to completions directory")
    print("  afdko completion fish > ~/.config/fish/completions/afdko.fish")
    print()
    print("  # PowerShell - add to profile")
    print('  afdko completion powershell | Out-String | Invoke-Expression')


def main() -> int:
    """Main entry point for completion command."""
    if len(sys.argv) < 2 or sys.argv[1] in ('-h', '--help', 'help'):
        print_help()
        return 0

    shell = sys.argv[1].lower()

    if shell not in SUPPORTED_SHELLS:
        print(f"Error: Unsupported shell '{shell}'", file=sys.stderr)
        print(f"Supported shells: {', '.join(SUPPORTED_SHELLS)}",
              file=sys.stderr)
        print("Run 'afdko completion --help' for more information.",
              file=sys.stderr)
        return 1

    # Get commands and generate completion script
    commands = get_commands()

    if shell == 'bash':
        script = generate_bash(commands)
    elif shell == 'zsh':
        script = generate_zsh(commands)
    elif shell == 'fish':
        script = generate_fish(commands)
    elif shell == 'powershell':
        script = generate_powershell(commands)
    else:
        # Should never reach here due to validation above
        return 1

    print(script, end='')
    return 0


if __name__ == '__main__':
    sys.exit(main())



"""
Deprecation wrappers for standalone command scripts.

These wrappers emit warnings before delegating to the actual
implementation. Commands are classified as "early" or "standard"
deprecation based on usage patterns.

Environment Variable:
    AFDKO_WRAPPER_MODE: Controls deprecation behavior
        - 'off' or unset: No warnings (current default)
        - 'warn': Show deprecation warnings (future default)
        - 'error': Treat deprecated wrappers as errors

    PYTHONWARNINGS=ignore: Also suppresses warnings (Python standard)

Timeline:
    Phase 1 (Current): Default='off', users opt-in with
                       AFDKO_WRAPPER_MODE=warn
    Phase 2 (Future):  Default='warn', users can silence with
                       AFDKO_WRAPPER_MODE=off
    Phase 3 (Near removal): Optionally default='error' for final
                            migration push
"""

import os
import sys
import warnings
from typing import Callable, NoReturn


# ============================================================================
# Configuration - Update these dates when releasing
# ============================================================================

# Current AFDKO version being released
CURRENT_VERSION = "4.1.0"
RELEASE_DATE = "2026-03-15"

# Date-based deprecation targets (hybrid approach)
# Early commands: 6 months after release
EARLY_REMOVAL_DATE = "2026-09-15"
EARLY_REMOVAL_TEXT = "an upcoming release after September 2026"

# Standard commands: 12 months after release, next major version
STANDARD_REMOVAL_DATE = "2027-03-15"
STANDARD_REMOVAL_VERSION = "5.0"
STANDARD_REMOVAL_TEXT = "the next major version after March 2027"

# Default wrapper mode (change this for different phases of rollout)
# Phase 1: 'off' - Users must opt-in to see warnings
# Phase 2: 'warn' - Show warnings by default (near removal dates)
# Phase 3: 'error' - Treat as errors for final migration (optional)
DEFAULT_WRAPPER_MODE = 'off'


# ============================================================================
# Command Classification
# ============================================================================

# Commands marked for early deprecation (6 months)
# Based on: /Users/skefiterum/src/docrepo/slides/afdko_command_planning.md
EARLY_COMMANDS = {
    # C++ commands
    'addfeatures',
    'detype1',
    'sfntdiff',
    'type1',
    # Python build commands
    'comparefamily',
    'otc2otf',
    'otf2otc',
    'ttfcomponentizer',
    'ttfdecomponentizer',
    # Python proofing commands
    'charplot',
    'digiplot',
    'fontplot',
    'fontplot2',
    'fontsetplot',
    'hintplot',
}

# All other commands use standard deprecation (12 months, next major version)
# These are the commonly used, high-impact commands


# ============================================================================
# Environment Variable Checks
# ============================================================================

def _get_wrapper_mode() -> str:
    """
    Get the current wrapper deprecation mode.

    Returns:
        'off', 'warn', or 'error'
    """
    # Python standard: PYTHONWARNINGS=ignore forces 'off'
    if os.environ.get('PYTHONWARNINGS') == 'ignore':
        return 'off'

    # Get AFDKO_WRAPPER_MODE, normalize to lowercase
    mode = os.environ.get('AFDKO_WRAPPER_MODE', '').lower().strip()

    # Handle common variations
    if mode in ('off', 'silent', 'disabled', 'no', 'false', '0'):
        return 'off'
    elif mode in ('warn', 'warning', 'warnings', 'yes', 'true', '1'):
        return 'warn'
    elif mode in ('error', 'strict', 'fail', '2'):
        return 'error'
    elif mode == '':
        # No environment variable set, use default
        return DEFAULT_WRAPPER_MODE
    else:
        # Invalid value, warn and use default
        print(
            f"Warning: Invalid AFDKO_WRAPPER_MODE='{mode}'. "
            f"Valid values: off, warn, error. "
            f"Using default: '{DEFAULT_WRAPPER_MODE}'",
            file=sys.stderr
        )
        return DEFAULT_WRAPPER_MODE


# ============================================================================
# Warning Messages
# ============================================================================

def _format_early_warning(command_name: str) -> str:
    """Format deprecation warning for early-removal commands."""
    return (
        f"DeprecationWarning: The '{command_name}' command wrapper is "
        f"deprecated and will be removed in {EARLY_REMOVAL_TEXT}.\n"
        f'The command should now be run as "afdko {command_name} '
        f'[options]".\n\n'
        f"To suppress this warning, set AFDKO_WRAPPER_MODE=off or "
        f"PYTHONWARNINGS=ignore.\n"
    )


def _format_standard_warning(command_name: str) -> str:
    """Format deprecation warning for standard-removal commands."""
    return (
        f"DeprecationWarning: The '{command_name}' command wrapper is "
        f"deprecated and will be removed in {STANDARD_REMOVAL_TEXT}.\n"
        f'The command should now be run as "afdko {command_name} '
        f'[options]".\n\n'
        f"To suppress this warning, set AFDKO_WRAPPER_MODE=off or "
        f"PYTHONWARNINGS=ignore.\n"
    )


def _emit_deprecation_warning(command_name: str) -> None:
    """
    Emit deprecation warning for a command wrapper.

    Args:
        command_name: Name of the deprecated command
    """
    # Check current mode
    mode = _get_wrapper_mode()

    # If mode is 'off', don't emit warnings
    if mode == 'off':
        return

    # Mode is 'warn' - emit the warning
    # (mode 'error' is handled separately in deprecated_command)

    # Determine if this is an early or standard deprecation
    is_early = command_name in EARLY_COMMANDS

    # Format the appropriate message
    if is_early:
        message = _format_early_warning(command_name)
    else:
        message = _format_standard_warning(command_name)

    # Print to stderr (always visible unless redirected)
    print(message, file=sys.stderr)

    # Also emit FutureWarning for Python contexts
    # (Use short message for Python warnings)
    removal = EARLY_REMOVAL_TEXT if is_early else STANDARD_REMOVAL_TEXT
    short_msg = (
        f"'{command_name}' wrapper is deprecated; "
        f"use 'afdko {command_name}' instead. "
        f"Removal: {removal}"
    )
    warnings.warn(short_msg, FutureWarning, stacklevel=3)


def _check_error_mode(command_name: str) -> None:
    """
    Check if wrapper should error instead of warning.

    Raises:
        SystemExit: If AFDKO_WRAPPER_MODE=error
    """
    mode = _get_wrapper_mode()

    if mode == 'error':
        is_early = command_name in EARLY_COMMANDS
        removal_text = (EARLY_REMOVAL_TEXT if is_early
                        else STANDARD_REMOVAL_TEXT)

        print(
            f"Error: The '{command_name}' wrapper is deprecated and will "
            f"be removed in {removal_text}.\n"
            f'The command should now be run as "afdko {command_name} '
            f'[options]".\n\n'
            f"To temporarily allow deprecated wrappers, set "
            f"AFDKO_WRAPPER_MODE=off or AFDKO_WRAPPER_MODE=warn.\n",
            file=sys.stderr
        )
        sys.exit(1)


# ============================================================================
# Wrapper Functions
# ============================================================================

def deprecated_command(command_name: str, target_func: Callable) -> Callable:
    """
    Create a deprecation wrapper for a command.

    Args:
        command_name: Name of the command (e.g., 'tx', 'makeotf')
        target_func: Actual implementation function

    Returns:
        Wrapped function that emits warnings before calling target
    """
    def wrapper(*args, **kwargs) -> NoReturn:
        # Check if we should error instead of warning
        _check_error_mode(command_name)

        # Emit deprecation warning
        _emit_deprecation_warning(command_name)

        # Call actual implementation
        return target_func(*args, **kwargs)

    wrapper.__name__ = target_func.__name__
    wrapper.__doc__ = (
        f"DEPRECATED: Use 'afdko {command_name}' instead.\n\n"
        f"{target_func.__doc__ or ''}"
    )
    return wrapper


# ============================================================================
# Individual Command Wrappers
# ============================================================================
# These are the entry points referenced in pyproject.toml

# C++ Commands
def tx_wrapper() -> NoReturn:
    """Deprecated wrapper for tx command."""
    from afdko._internal import tx
    return deprecated_command('tx', tx)()


def sfntedit_wrapper() -> NoReturn:
    """Deprecated wrapper for sfntedit command."""
    from afdko._internal import sfntedit
    return deprecated_command('sfntedit', sfntedit)()


def spot_wrapper() -> NoReturn:
    """Deprecated wrapper for spot command."""
    from afdko._internal import spot
    return deprecated_command('spot', spot)()


def addfeatures_wrapper() -> NoReturn:
    """Deprecated wrapper for addfeatures command (EARLY)."""
    from afdko._internal import addfeatures
    return deprecated_command('addfeatures', addfeatures)()


def detype1_wrapper() -> NoReturn:
    """Deprecated wrapper for detype1 command (EARLY)."""
    from afdko._internal import detype1
    return deprecated_command('detype1', detype1)()


def type1_wrapper() -> NoReturn:
    """Deprecated wrapper for type1 command (EARLY)."""
    from afdko._internal import type1
    return deprecated_command('type1', type1)()


def sfntdiff_wrapper() -> NoReturn:
    """Deprecated wrapper for sfntdiff command (EARLY)."""
    from afdko._internal import sfntdiff
    return deprecated_command('sfntdiff', sfntdiff)()


def mergefonts_wrapper() -> NoReturn:
    """Deprecated wrapper for mergefonts command."""
    from afdko._internal import mergefonts
    return deprecated_command('mergefonts', mergefonts)()


def rotatefont_wrapper() -> NoReturn:
    """Deprecated wrapper for rotatefont command."""
    from afdko._internal import rotatefont
    return deprecated_command('rotatefont', rotatefont)()


# Python Commands - Build & Processing
def makeotf_wrapper() -> NoReturn:
    """Deprecated wrapper for makeotf command."""
    from afdko.makeotf import main
    return deprecated_command('makeotf', main)()


def buildcff2vf_wrapper() -> NoReturn:
    """Deprecated wrapper for buildcff2vf command."""
    from afdko.buildcff2vf import main
    return deprecated_command('buildcff2vf', main)()


def buildmasterotfs_wrapper() -> NoReturn:
    """Deprecated wrapper for buildmasterotfs command."""
    from afdko.buildmasterotfs import main
    return deprecated_command('buildmasterotfs', main)()


def makeinstancesufo_wrapper() -> NoReturn:
    """Deprecated wrapper for makeinstancesufo command."""
    from afdko.makeinstancesufo import main
    return deprecated_command('makeinstancesufo', main)()


def checkoutlinesufo_wrapper() -> NoReturn:
    """Deprecated wrapper for checkoutlinesufo command."""
    from afdko.checkoutlinesufo import main
    return deprecated_command('checkoutlinesufo', main)()


def comparefamily_wrapper() -> NoReturn:
    """Deprecated wrapper for comparefamily command (EARLY)."""
    from afdko.comparefamily import main
    return deprecated_command('comparefamily', main)()


def otc2otf_wrapper() -> NoReturn:
    """Deprecated wrapper for otc2otf command (EARLY)."""
    from afdko.otc2otf import main
    return deprecated_command('otc2otf', main)()


def otf2otc_wrapper() -> NoReturn:
    """Deprecated wrapper for otf2otc command (EARLY)."""
    from afdko.otf2otc import main
    return deprecated_command('otf2otc', main)()


def otf2ttf_wrapper() -> NoReturn:
    """Deprecated wrapper for otf2ttf command."""
    from afdko.otf2ttf import main
    return deprecated_command('otf2ttf', main)()


def ttfcomponentizer_wrapper() -> NoReturn:
    """Deprecated wrapper for ttfcomponentizer command (EARLY)."""
    from afdko.ttfcomponentizer import main
    return deprecated_command('ttfcomponentizer', main)()


def ttfdecomponentizer_wrapper() -> NoReturn:
    """Deprecated wrapper for ttfdecomponentizer command (EARLY)."""
    from afdko.ttfdecomponentizer import main
    return deprecated_command('ttfdecomponentizer', main)()


def ttxn_wrapper() -> NoReturn:
    """Deprecated wrapper for ttxn command."""
    from afdko.ttxn import main
    return deprecated_command('ttxn', main)()


# Python Commands - Proofing (all EARLY)
def charplot_wrapper() -> NoReturn:
    """Deprecated wrapper for charplot command (EARLY)."""
    from afdko.proofpdf import charplot
    return deprecated_command('charplot', charplot)()


def digiplot_wrapper() -> NoReturn:
    """Deprecated wrapper for digiplot command (EARLY)."""
    from afdko.proofpdf import digiplot
    return deprecated_command('digiplot', digiplot)()


def fontplot_wrapper() -> NoReturn:
    """Deprecated wrapper for fontplot command (EARLY)."""
    from afdko.proofpdf import fontplot
    return deprecated_command('fontplot', fontplot)()


def fontplot2_wrapper() -> NoReturn:
    """Deprecated wrapper for fontplot2 command (EARLY)."""
    from afdko.proofpdf import fontplot2
    return deprecated_command('fontplot2', fontplot2)()


def fontsetplot_wrapper() -> NoReturn:
    """Deprecated wrapper for fontsetplot command (EARLY)."""
    from afdko.proofpdf import fontsetplot
    return deprecated_command('fontsetplot', fontsetplot)()


def hintplot_wrapper() -> NoReturn:
    """Deprecated wrapper for hintplot command (EARLY)."""
    from afdko.proofpdf import hintplot
    return deprecated_command('hintplot', hintplot)()


def waterfallplot_wrapper() -> NoReturn:
    """Deprecated wrapper for waterfallplot command."""
    from afdko.proofpdf import waterfallplot
    return deprecated_command('waterfallplot', waterfallplot)()


# Python Commands - Hinting
def otfautohint_wrapper() -> NoReturn:
    """Deprecated wrapper for otfautohint command."""
    from afdko.otfautohint.__main__ import main
    return deprecated_command('otfautohint', main)()


def otfstemhist_wrapper() -> NoReturn:
    """Deprecated wrapper for otfstemhist command."""
    from afdko.otfautohint.__main__ import stemhist
    return deprecated_command('otfstemhist', stemhist)()


# ============================================================================
# Special Transition Wrappers (Always Show Message)
# ============================================================================

def makeotfexe_wrapper() -> NoReturn:
    """Transition message for makeotfexe -> addfeatures."""
    import sys

    message = (
        "The 'makeotfexe' command has been replaced by 'addfeatures'.\n\n"
        "makeotfexe took a Type 1 font, converted it to a CFF table, and "
        "compiled features into OpenType tables. addfeatures takes a CFF "
        "table as input and compiles the features. The 'makeotf' wrapper "
        "handles this change automatically, so most users won't need to "
        "adjust their workflows.\n\n"
        "If you were calling makeotfexe directly, you may now need to "
        "build a CFF table first using 'afdko tx', which has been "
        "enhanced with some features that were previously exclusive to "
        "makeotfexe, including support for GlyphOrderAndAliasDB files.\n\n"
        "Examples:\n"
        "  Old: makeotfexe -f features.fea font.pfa\n"
        "  New: afdko tx -cff +b font.pfa temp.cff && "
        "afdko addfeatures -f features.fea temp.cff\n\n"
        "  Or use the makeotf command: afdko makeotf -f font.pfa "
        "-ff features.fea\n\n"
        "The makeotfexe wrapper will be removed in the next major "
        "version after March 2027.\n"
    )

    print(message, file=sys.stderr)
    sys.exit(1)



# Copyright 2022 Adobe. All rights reserved.

# Portions of the code in removeOverlap are adapted from the MIT-licensed
# booleanOperations python package, located at
# https://github.com/typemytype/booleanOperations
# parts of which are also imported

from booleanOperations.flatten import InputContour, OutputContour
from booleanOperations.booleanOperationManager import clipExecute
from booleanOperations.booleanGlyph import BooleanGlyph
from fontTools.pens.pointPen import PointToSegmentPen
from .glyphData import glyphData
import logging

log = logging.getLogger(__name__)


def removeOverlap(glyph: glyphData) -> glyphData | None:
    """
    If the glyphData 'glyph' object has overlap, create a new glyphData
    object with the overlap removed and return it.  If it has no overlap
    return None
    """
    bg = BooleanGlyph()
    pointPen = bg.getPointPen()
    pointPen.copyContourData = True
    glyph.drawPoints(pointPen)

    inputContours = [InputContour(c) for c in bg.contours if c and len(c) > 1]
    inputFlats = [c.originalFlat for c in inputContours]
    it = 0
    while True:
        resultFlats = clipExecute(inputFlats, [], "union",
                                  subjectFillType="nonZero",
                                  clipFillType="nonZero")
        iPoints = set()
        for c in inputFlats:
            iPoints.update(c)
        rPoints = set()
        for c in resultFlats:
            rPoints.update(c)
        intersects = iPoints - rPoints
        if not intersects or it >= 8:
            break
        inputFlats = resultFlats
        it += 1
    if not intersects and it == 0:
        return None
    elif intersects:
        log.warning("Could not eliminate all overlap within 8 iterations")

    outputContours = [OutputContour(c) for c in resultFlats]
    for ic in inputContours:
        for oc in outputContours:
            if oc.final:
                continue
            if oc.reCurveFromEntireInputContour(ic):
                break
    for oc in outputContours:
        oc.reCurveSubSegments(inputContours)

    dest = glyphData(roundCoords=False, name=glyph.name)
    pointPen = PointToSegmentPen(dest)
    for oc in outputContours:
        oc.drawPoints(pointPen)
    return dest
